import sys
import numpy as np
import pandas as pd
import os
import time
import datetime
from tensorflow import keras
import matplotlib.pyplot as plt
from sklearn.model_selection import train_test_split
import opensim as osim

current_dir = os.path.dirname(os.path.abspath(__file__))
parent_dir = os.path.abspath(os.path.join(current_dir, '..', '..'))
sys.path.append(parent_dir)
from program.sub.set_initial_state_from_sto import set_initial_state_from_sto
from program.sub.Est_Lag import calculate_estimation_lag
from program.sub.ses_tuning import update_parameters
from program.sub.create_target_trajectory import Target_Angle
from program.sub.get_text import get_text
from program.sub.custom_scalers import CustomScaler, CustomInverseScaler
from program.sub.pid_controller import PID_controller
from program.sub.pid_ann_controller import PID_ANN_controller
from program.sub.fuzzy_controller import create_fuzzy_controller, Fuzzy_controller
from program.sub.recruitment_characteristics import sigmoid_function, logarithmic_function, quadratic_function, cubic_function, first_order_delay_system, tanh, sqrt, sine_squared
from program.sub.normalize_list import normalize_list

def control_main(control_condition, target_muscle, target_pattern, quiet_time, end_time, excitation_BIC_max, excitation_TRI_max, excitation_BIC_min, excitation_TRI_min, zero_point, amplitude, frequency, model_file_name, model_folder_name, IniSta_file_name, IniSta_folder_name, result_folder_name, iteration, total_iteration, training_history, TIME_STEP, params_BIC, params_TRI, Flags, target_period, recruitment_paras, ANN_forder_name, recruitment_pattern):

    # 諸変数
    start_time = 0.0
    ramp_start = quiet_time # Ramp開始
    ramp_end = end_time # Ramp終了
    step_start = quiet_time # Step開始
    step_end = end_time # Step終了
    ramp2step = 1 + quiet_time # Ramp→Stepの切り替え時間(Ramp+Step)
    switch_time = (end_time / 2) - quiet_time
    sin_start = 7
    excitation_BIC = excitation_BIC_min
    excitation_TRI = excitation_TRI_min
    excitation_max_minus_min_BIC = excitation_BIC_max - excitation_BIC_min
    excitation_max_minus_min_TRI = excitation_TRI_max - excitation_TRI_min
    excitation_PID_BIC = 0
    excitation_ANN_BIC = 0
    Power_Ratio_ANN_BIC = 0
    Power_Ratio_ANN_TRI = 0
    PI = np.pi
    initial_sto_file = "{}/{}".format(IniSta_folder_name, IniSta_file_name)
    INITIAL_ANGLE_RAD = np.radians(float(get_text(initial_sto_file, 8, 31, 44)))
    global T1, T2, T3
    T1 = 0
    T2 = 0
    T3 = 0
    for_sec_index = int(4 / TIME_STEP)  # 3秒に相当するインデックス
    sh_REC, xc = recruitment_paras
    interval_steps = int(0.25 / TIME_STEP / 5)

    # flag
    bug_flag = False
    iteration_flag, plot_flag, ANN_training_flag, excitation_frequency_flag, SES_flag, recruitment_flag, delay_flag = Flags
    excitation_frequency = 20
    downsamplerate = int(1/excitation_frequency/TIME_STEP)

    # 制御器の名前
    if control_condition == 0:
        controller_name = "Step"
        TargetFlag = False
    elif control_condition == 1:
        controller_name = "Ramp"
        TargetFlag = False
    elif control_condition == 2:
        controller_name = "PID"
        TargetFlag = True
    elif control_condition == 3:
        controller_name = "Ramp2Step"
        TargetFlag = False
    elif control_condition == 5:
        controller_name = "Sin_Wave"
        TargetFlag = False
    elif control_condition == 6:
        controller_name = "FEL"
        TargetFlag = True
    elif control_condition == 7:
        controller_name = "Triangle_Wage"
        TargetFlag = False
    elif control_condition == 8:
        controller_name = "Fuzzy"
        TargetFlag = True
        error_mfs, delta_v_values, rules = create_fuzzy_controller()
        excitation_BIC = excitation_BIC_min

    # モデル初期設定
    osimmodel = osim.Model("{}/{}".format(model_folder_name, model_file_name))
    state = osimmodel.initSystem()
    state = set_initial_state_from_sto(osimmodel, initial_sto_file)
    BIClong = osimmodel.getMuscles().get("BIClong")
    BICshort = osimmodel.getMuscles().get("BICshort")
    TRIlong = osimmodel.getMuscles().get("TRIlong")
    TRIlat = osimmodel.getMuscles().get("TRIlat")
    TRImed = osimmodel.getMuscles().get("TRImed")
    Bra = osimmodel.getMuscles().get("BRA")

    methods = [attr for attr in BIClong.__dict__ if callable(getattr(BIClong, attr))]
    print(methods)

    # パラメータ格納用変数
    n_samples = int(end_time/TIME_STEP) + 1
    Time = np.full(n_samples, np.nan)
    Excitation_BIC = np.full(n_samples, np.nan)
    Excitation_BIC_norm = np.full(n_samples, np.nan)
    Excitation_ANN_BIC = np.full(n_samples, np.nan)
    Excitation_ANN_BIC_norm = np.full(n_samples, np.nan)
    Excitation_PID_BIC = np.full(n_samples, np.nan)
    Excitation_PID_BIC_norm = np.full(n_samples, np.nan)
    TrainingData_Excitation_BIC = np.full(n_samples, np.nan)
    TrainingData_Excitation_BIC_norm = np.full(n_samples, np.nan)
    Excitation_TRI = np.full(n_samples, np.nan)
    Excitation_TRI_norm = np.full(n_samples, np.nan)
    Excitation_ANN_TRI = np.full(n_samples, np.nan)
    Excitation_ANN_TRI_norm = np.full(n_samples, np.nan)
    Excitation_PID_TRI = np.full(n_samples, np.nan)
    Excitation_PID_TRI_norm = np.full(n_samples, np.nan)
    TrainingData_Excitation_TRI = np.full(n_samples, np.nan)
    TrainingData_Excitation_TRI_norm = np.full(n_samples, np.nan)
    ObservedAngle_rad = np.full(n_samples, np.nan)
    ObservedAngle_degree = np.full(n_samples, np.nan)
    TargetAngle_degree = np.full(n_samples, np.nan)
    ObservedAngularVelocity_rad = np.full(n_samples, np.nan)
    ObservedAngularVelocity_degree = np.full(n_samples, np.nan)
    TargetAngularVelocity_degree = np.full(n_samples, np.nan)
    ObservedAngularAcceleration_degree = np.full(n_samples, np.nan)
    TargetAngularAcceleration_degree = np.full(n_samples, np.nan)
    Error_rad = np.full(n_samples, np.nan)
    Error_degree = np.full(n_samples, np.nan)
    Mae_col = np.full(n_samples, np.nan)
    Rmse_col = np.full(n_samples, np.nan)
    Est_Lag_col = np.full(n_samples, np.nan)
    end_time_col = np.full(n_samples, np.nan)
    time_step_col = np.full(n_samples, np.nan)
    elapsed_time_col = np.full(n_samples, np.nan)
    Power_Ratio_ANN_BIC_col = np.full(n_samples, np.nan)
    Power_Ratio_ANN_TRI_col = np.full(n_samples, np.nan)
    T_col = np.full(n_samples, np.nan)
    costs_col = np.full(n_samples, np.nan)
    kp_col = np.full(n_samples, np.nan)
    ki_col = np.full(n_samples, np.nan)
    kd_col = np.full(n_samples, np.nan)
    kp_hat_col = np.full(n_samples, np.nan)
    ki_hat_col = np.full(n_samples, np.nan)
    kd_hat_col = np.full(n_samples, np.nan)
    noise_kp_col = np.full(n_samples, np.nan)
    noise_ki_col = np.full(n_samples, np.nan)
    noise_kd_col = np.full(n_samples, np.nan)
    f_x_col = np.full(n_samples, np.nan)

    Excitation_B = np.full(n_samples, np.nan)
    Activation_B = np.full(n_samples, np.nan)
    FiberLength_B = np.full(n_samples, np.nan)
    PennationAngle_B = np.full(n_samples, np.nan)
    CosPennationAngle_B = np.full(n_samples, np.nan)
    TendonLength_B = np.full(n_samples, np.nan)
    NormalizedFiberLength_B = np.full(n_samples, np.nan)
    FiberLengthAlongTendon_B = np.full(n_samples, np.nan)
    TendonStrain_B = np.full(n_samples, np.nan)
    PassiveForceMultiplier_B = np.full(n_samples, np.nan)
    ActiveForceLengthMultiplier_B = np.full(n_samples, np.nan)
    FiberVelocity_B = np.full(n_samples, np.nan)
    NormalizedFiberVelocity_B = np.full(n_samples, np.nan)
    FiberVelocityAlongTendon_B = np.full(n_samples, np.nan)
    TendonVelocity_B = np.full(n_samples, np.nan)
    ForceVelocityMultiplier_B = np.full(n_samples, np.nan)
    PennationAngularVelocity_B = np.full(n_samples, np.nan)
    FiberForce_B = np.full(n_samples, np.nan)
    FiberForceAlongTendon_B = np.full(n_samples, np.nan)
    ActiveFiberForce_B = np.full(n_samples, np.nan)
    PassiveFiberForce_B = np.full(n_samples, np.nan)
    ActiveFiberForceAlongTendon_B = np.full(n_samples, np.nan)
    PassiveFiberForceAlongTendon_B = np.full(n_samples, np.nan)
    TendonForce_B = np.full(n_samples, np.nan)
    FiberStiffness_B = np.full(n_samples, np.nan)
    FiberStiffnessAlongTendon_B = np.full(n_samples, np.nan)
    TendonStiffness_B = np.full(n_samples, np.nan)
    MuscleStiffness_B = np.full(n_samples, np.nan)
    FiberActivePower_B = np.full(n_samples, np.nan)
    FiberPassivePower_B = np.full(n_samples, np.nan)
    TendonPower_B = np.full(n_samples, np.nan)
    MusclePower_B = np.full(n_samples, np.nan)
    Stress_B = np.full(n_samples, np.nan)
    FiberPotentialEnergy_B = np.full(n_samples, np.nan)
    TendonPotentialEnergy_B = np.full(n_samples, np.nan)
    MusclePotentialEnergy_B = np.full(n_samples, np.nan)

    Excitation_T = np.full(n_samples, np.nan)
    Activation_T = np.full(n_samples, np.nan)
    FiberLength_T = np.full(n_samples, np.nan)
    PennationAngle_T = np.full(n_samples, np.nan)
    CosPennationAngle_T = np.full(n_samples, np.nan)
    TendonLength_T = np.full(n_samples, np.nan)
    NormalizedFiberLength_T = np.full(n_samples, np.nan)
    FiberLengthAlongTendon_T = np.full(n_samples, np.nan)
    TendonStrain_T= np.full(n_samples, np.nan)
    PassiveForceMultiplier_T = np.full(n_samples, np.nan)
    ActiveForceLengthMultiplier_T = np.full(n_samples, np.nan)
    FiberVelocity_T = np.full(n_samples, np.nan)
    NormalizedFiberVelocity_T = np.full(n_samples, np.nan)
    FiberVelocityAlongTendon_T = np.full(n_samples, np.nan)
    TendonVelocity_T = np.full(n_samples, np.nan)
    ForceVelocityMultiplier_T = np.full(n_samples, np.nan)
    PennationAngularVelocity_T = np.full(n_samples, np.nan)
    FiberForce_T = np.full(n_samples, np.nan)
    FiberForceAlongTendon_T = np.full(n_samples, np.nan)
    ActiveFiberForce_T = np.full(n_samples, np.nan)
    PassiveFiberForce_T = np.full(n_samples, np.nan)
    ActiveFiberForceAlongTendon_T = np.full(n_samples, np.nan)
    PassiveFiberForceAlongTendon_T = np.full(n_samples, np.nan)
    TendonForce_T = np.full(n_samples, np.nan)
    FiberStiffness_T = np.full(n_samples, np.nan)
    FiberStiffnessAlongTendon_T = np.full(n_samples, np.nan)
    TendonStiffness_T = np.full(n_samples, np.nan)
    MuscleStiffness_T = np.full(n_samples, np.nan)
    FiberActivePower_T = np.full(n_samples, np.nan)
    FiberPassivePower_T = np.full(n_samples, np.nan)
    TendonPower_T = np.full(n_samples, np.nan)
    MusclePower_T = np.full(n_samples, np.nan)
    Stress_T = np.full(n_samples, np.nan)
    FiberPotentialEnergy_T = np.full(n_samples, np.nan)
    TendonPotentialEnergy_T = np.full(n_samples, np.nan)
    MusclePotentialEnergy_T = np.full(n_samples, np.nan)

    Excitation_BIC_input = np.full(n_samples, np.nan)
    Excitation_TRI_input = np.full(n_samples, np.nan)

    delay_time_BIC = 0.2
    delay_time_TRI = 0.2
    n_delay_samples_BIC = int(delay_time_BIC/TIME_STEP)
    n_delay_samples_TRI = int(delay_time_TRI/TIME_STEP)
    delay_BIC = list(np.full(n_delay_samples_BIC, excitation_BIC_min))
    delay_TRI = list(np.full(n_delay_samples_TRI, excitation_TRI_min))

    # 初期のPIDパラメーター設定
    kp_col[0] = params_BIC[0]
    ki_col[0] = params_BIC[1]
    kd_col[0] = params_BIC[2]
    kp_hat_col[0] = params_BIC[0]
    ki_hat_col[0] = params_BIC[1]
    kd_hat_col[0] = params_BIC[2]
    pid_parameters = [params_BIC]
    pid_parameters_hat = [params_BIC]
    noises = [[0,0,0]]
    noise_kp_col[0] = 0
    noise_ki_col[0] = 0
    noise_kd_col[0] = 0

    # 追従実験時の目標角度波形
    if TargetFlag:
        print("目標軌道作成中")
        time_array = np.arange(start_time, end_time + TIME_STEP + 0.25, TIME_STEP)
        target_angles, target_velocities, target_accelerations = Target_Angle(target_pattern, target_period, time_array, INITIAL_ANGLE_RAD, TIME_STEP, end_time)
    # ANNによる予測
    if control_condition == 6:  
        print("ANN(FF)による予測中")
        # モデルの読み込み
        ANNmodel = keras.models.load_model(f'{model_folder_name}/{ANN_forder_name}/pre_learned_model_{iteration-1}.keras')
        inputs = []
        for i in range(len(target_angles) - interval_steps * 5):
            # データを間隔を空けて取得
            angle = np.degrees([target_angles[i + interval_steps * j] for j in range(6)])
            angular_velocity = np.degrees([target_velocities[i + interval_steps * j] for j in range(6)])
            angular_acceleration = np.degrees([target_accelerations[i + interval_steps * j] for j in range(6)])
            
            # 入力データの正規化
            custom_scaler_inputs_angle = CustomScaler(10, 90)
            custom_scaler_inputs_velocity = CustomScaler(-68.3, 68.3)
            custom_scaler_inputs_acceleration = CustomScaler(-180, 180)
            
            # 正規化
            angle = custom_scaler_inputs_angle.fit_transform(angle)
            angular_velocity = custom_scaler_inputs_velocity.fit_transform(angular_velocity)
            angular_acceleration = custom_scaler_inputs_acceleration.fit_transform(angular_acceleration)

            # 入力データの統合
            input_data = np.hstack([angle, angular_velocity, angular_acceleration])
            inputs.append(input_data)
        inputs = np.array(inputs)
        # 値の予測
        FF_output = ANNmodel.predict(inputs)
        FF_output_BIC = np.full(n_samples, np.nan)
        FF_output_TRI = np.full(n_samples, np.nan)
        for i in range (n_samples):
            FF_output_BIC[i] = FF_output[i][0]
            FF_output_TRI[i] = FF_output[i][1]
        output_scaler_BIC = CustomInverseScaler(-0.1, excitation_max_minus_min_BIC)
        output_scaler_TRI = CustomInverseScaler(-0.1, excitation_max_minus_min_TRI)
        # 予測結果を元のスケールに逆変換（出力データ用スケーラーを使用）
        FF_output_BIC = output_scaler_BIC.inverse_transform(FF_output_BIC)
        FF_output_TRI = output_scaler_TRI.inverse_transform(FF_output_TRI)
            
    # 制御器を仮作成
    # BIClong用
    controller_BIClong = osim.PrescribedController()
    controller_BIClong.setName("controller_BIClong")
    excitation_function_BIClong = osim.Constant(excitation_BIC_min)
    controller_BIClong.addActuator(BIClong)
    controller_BIClong.prescribeControlForActuator(BIClong.getName(), excitation_function_BIClong)
    osimmodel.addController(controller_BIClong)
    # BICshort用
    controller_BICshort = osim.PrescribedController()
    controller_BICshort.setName("controller_BICshort")
    excitation_function_BICshort = osim.Constant(excitation_BIC_min)
    controller_BICshort.addActuator(BICshort)
    controller_BICshort.prescribeControlForActuator(BICshort.getName(), excitation_function_BICshort)
    osimmodel.addController(controller_BICshort)
    # TRIlong用
    controller_TRIlong = osim.PrescribedController()
    controller_TRIlong.setName("controller_TRIlong")
    excitation_function_TRIlong = osim.Constant(excitation_TRI_min)
    controller_TRIlong.addActuator(TRIlong)
    controller_TRIlong.prescribeControlForActuator(TRIlong.getName(), excitation_function_TRIlong)
    osimmodel.addController(controller_TRIlong)
    # TRIlat用
    controller_TRIlat = osim.PrescribedController()
    controller_TRIlat.setName("controller_TRIlat")
    excitation_function_TRIlat = osim.Constant(excitation_TRI_min)
    controller_TRIlat.addActuator(TRIlat)
    controller_TRIlat.prescribeControlForActuator(TRIlat.getName(), excitation_function_TRIlat)
    osimmodel.addController(controller_TRIlat)
    # TRImed用
    controller_TRImed = osim.PrescribedController()
    controller_TRImed.setName("controller_TRImed")
    excitation_function_TRImed = osim.Constant(excitation_TRI_min)
    controller_TRImed.addActuator(TRImed)
    controller_TRImed.prescribeControlForActuator(TRImed.getName(), excitation_function_TRImed)
    osimmodel.addController(controller_TRImed)
    # Bra用
    controller_Bra = osim.PrescribedController()
    controller_Bra.setName("controller_Bra")
    excitation_function_Bra = osim.Constant(excitation_BIC_min)
    controller_Bra.addActuator(Bra)
    controller_Bra.prescribeControlForActuator(Bra.getName(), excitation_function_Bra)
    osimmodel.addController(controller_Bra)

    # CoordinateLimitForceを設定(肘関節の可動域設定)
    elbow_limit_force = osim.CoordinateLimitForce("r_elbow_flex", 150, 5, 1.5, 5, 0.1, 0.001)
    osimmodel.addForce(elbow_limit_force)

    # モデルに各種設定を反映
    state = osimmodel.initSystem()

    # シミュレーションオプション
    manager = osim.Manager(osimmodel)
    manager.setIntegratorMethod(osim.Manager.IntegratorMethod_RungeKuttaMerson)
    manager.setIntegratorMaximumStepSize(TIME_STEP)
    manager.setIntegratorMinimumStepSize(TIME_STEP)
    manager.initialize(state)
    state.setTime(start_time)

    # シミュレーション開始時の初期設定
    TIME_SIMULATION_START = time.time()
    if TargetFlag:
        last_error_BIC = target_angles[0] - INITIAL_ANGLE_RAD
        last_integral_error_BIC = 0
        last_error_TRI = target_angles[0] - INITIAL_ANGLE_RAD
        last_integral_error_TRI = 0

    # グラフ描画用設定
    if plot_flag:
        plt.ion()
        fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(10, 8))  # 2つのプロット領域を作成
        # シミュレーション中のデータ格納用リスト
        current_times = []
        state_values1 = []
        state_values2 = []
        state_values3 = []
        if TargetFlag:
            state_values4 = []

    # シミュレーション実行部分
    # 保存されるデータは時刻iにおいて積分後のデータ
    for i in range (n_samples):
        current_time = TIME_STEP * i
        # 初期時刻(t=0.0000)時の動作
        if current_time == 0:
            # 初期状態保存
            Time[i] = current_time
            Excitation_BIC[i] = excitation_BIC_min
            Excitation_TRI[i] = excitation_TRI_min
            Excitation_BIC_input[i] = excitation_BIC_min
            Excitation_TRI_input[i] = excitation_TRI_min
            ObservedAngle_rad[i] = osimmodel.getStateVariableValues(state).get(2)
            ObservedAngularVelocity_rad[i] = osimmodel.getStateVariableValues(state).get(3)
            if TargetFlag:
                if control_condition == 6:
                    Excitation_PID_BIC[i] = 0
                    Excitation_ANN_BIC[i] = 0
                    TrainingData_Excitation_BIC[i] = 0
                    Excitation_PID_TRI[i] = 0
                    Excitation_ANN_TRI[i] = 0
                    TrainingData_Excitation_TRI[i] = 0
                Error_rad[i] = target_angles[0] - INITIAL_ANGLE_RAD
            if not iteration_flag:
                print("{}シミュレーション中:{}/{}".format(controller_name, round(current_time,2), end_time))
            elif iteration_flag:
                print("{}回/{}回 {}シミュレーション中:{}/{}".format(iteration, total_iteration, controller_name, round(current_time,2), end_time))
            
            

            # グラフ描画用
            if plot_flag:
                current_times.append(current_time)
                state_values1.append(np.degrees(ObservedAngle_rad[i]))
                state_values2.append(Excitation_BIC[i])
                state_values3.append(Excitation_TRI[i])
                if TargetFlag:
                    state_values4.append(np.degrees(target_angles[i]))
            continue
        # 制御器の設定
        if control_condition == 0: #Step
            if current_time < step_start:
                excitation_BIC = excitation_BIC_min
                excitation_TRI = excitation_TRI_min
            elif current_time <= end_time:
                excitation_BIC = excitation_BIC_max
                excitation_TRI = excitation_TRI_max
            else:
                excitation_BIC = excitation_BIC_min
                excitation_TRI = excitation_TRI_min
        if control_condition == 1: #Ramp
            if current_time < ramp_start:
                excitation_BIC = excitation_BIC_min
                excitation_TRI = excitation_TRI_min
            elif current_time <= ramp_end:
                excitation_BIC = excitation_BIC_min  + (current_time - ramp_start) * (excitation_BIC_max - excitation_BIC_min) / (ramp_end - ramp_start)
                excitation_TRI = excitation_TRI_min  + (current_time - ramp_start) * (excitation_TRI_max - excitation_TRI_min) / (ramp_end - ramp_start)
            else:
                excitation_BIC = excitation_BIC_min
                excitation_TRI = excitation_TRI_min
        if control_condition == 2: #PID
            target_angle = target_angles[i]

            # SESでのパラメータ更新
            if SES_flag and current_time > 6 and i % int(5 / TIME_STEP) == 1:
                n_period = int(current_time / 5)
                i_cutted1 = int(5 / TIME_STEP)
                target_angles_cutted = target_angles[1 + i_cutted1 * (n_period - 1) : i_cutted1 * n_period]
                ObservedAngle_rad_cutted = ObservedAngle_rad[1 + i_cutted1 * (n_period - 1) : i_cutted1 * n_period]
                theta_n_hat ,params_BIC, sin_v, cost, f_x = update_parameters(n_period-1, pid_parameters, pid_parameters_hat, noises, costs_col, target_angles_cutted, ObservedAngle_rad_cutted)
                pid_parameters_hat.append(theta_n_hat)
                pid_parameters.append(params_BIC)
                costs_col[n_period-2] = cost
                kp_col[n_period-1] = params_BIC[0]
                ki_col[n_period-1] = params_BIC[1]
                kd_col[n_period-1] = params_BIC[2]
                kp_hat_col[n_period-1] = theta_n_hat[0]
                ki_hat_col[n_period-1] = theta_n_hat[1]
                kd_hat_col[n_period-1] = theta_n_hat[2]
                noises.append(sin_v)
                noise_kp_col[n_period-1] = sin_v[0]
                noise_ki_col[n_period-1] = sin_v[1]
                noise_kd_col[n_period-1] = sin_v[2]
                f_x_col[n_period-2] = f_x

            if not excitation_frequency_flag:
                excitation_BIC, last_error_BIC, last_integral_error_BIC = PID_controller(target_angle, ObservedAngle_rad[i-1], last_error_BIC, last_integral_error_BIC, params_BIC, TIME_STEP, excitation_BIC_max, excitation_BIC_min)
                excitation_TRI, last_error_TRI, last_integral_error_TRI = PID_controller(target_angle, ObservedAngle_rad[i-1], last_error_TRI, last_integral_error_TRI, params_TRI, TIME_STEP, excitation_TRI_max, excitation_TRI_min)
            else:
                if i % int(1/excitation_frequency/TIME_STEP) == 1:
                    excitation_BIC, last_error_BIC, last_integral_error_BIC = PID_controller(target_angle, ObservedAngle_rad[i-1], last_error_BIC, last_integral_error_BIC, params_BIC, (1 / excitation_frequency), excitation_BIC_max, excitation_BIC_min)
                    excitation_TRI, last_error_TRI, last_integral_error_TRI = PID_controller(target_angle, ObservedAngle_rad[i-1], last_error_TRI, last_integral_error_TRI, params_TRI, (1 / excitation_frequency), excitation_TRI_max, excitation_TRI_min)
        if control_condition == 3: #Ramp2Step
            if current_time < ramp_start:
                excitation_BIC = excitation_BIC_min 
            elif current_time <= ramp2step:
                excitation_BIC = excitation_BIC_min  + (current_time - ramp_start) * (excitation_BIC_max - excitation_BIC_min) / (ramp2step - ramp_start)
            else:
                excitation_BIC = excitation_BIC_max
        if control_condition == 5: # Sin_Wave
            sec = end_time - sin_start
            t = np.arange(0, sec, TIME_STEP) #サンプリング点の生成
            y = np.sin(2*PI*frequency*t) # 正弦波の生成
            if current_time < ramp_start:
                excitation_BIC = excitation_BIC_min 
            elif current_time <= ramp2step:
                excitation_BIC = excitation_BIC_min  + (current_time - ramp_start) * (zero_point - excitation_BIC_min) / (ramp2step - ramp_start)
            elif current_time <= sin_start:
                excitation_BIC = zero_point
            else:
                excitation_BIC = zero_point + amplitude * y[i - int(7 / TIME_STEP) - 1]
        if control_condition == 6: # Feedback-Error-Learning controller
            target_angle = target_angles[i]
            excitation_ANN_BIC = FF_output_BIC[i]
            excitation_ANN_TRI = FF_output_TRI[i]
            excitation_PID_BIC, last_error_BIC, last_integral_error_BIC = PID_ANN_controller(target_angle, ObservedAngle_rad[i-1], last_error_BIC, last_integral_error_BIC, excitation_ANN_BIC, params_BIC, TIME_STEP, excitation_max_minus_min_BIC)
            excitation_PID_TRI, last_error_TRI, last_integral_error_TRI = PID_ANN_controller(target_angle, ObservedAngle_rad[i-1], last_error_TRI, last_integral_error_TRI, excitation_ANN_TRI, params_TRI, TIME_STEP, excitation_max_minus_min_TRI)
            trainingdata_excitation_BIC = excitation_PID_BIC + excitation_ANN_BIC
            trainingdata_excitation_TRI = excitation_PID_TRI + excitation_ANN_TRI
            excitation_BIC = excitation_PID_BIC + excitation_ANN_BIC + excitation_BIC_min # Offset加算
            excitation_TRI = excitation_PID_TRI + excitation_ANN_TRI + excitation_TRI_min # Offset加算
            # Limiter
            if excitation_BIC > excitation_BIC_max:
                excitation_BIC = excitation_BIC_max
            elif excitation_BIC < excitation_BIC_min:
                excitation_BIC = excitation_BIC_min
            if excitation_TRI > excitation_TRI_max:
                excitation_TRI = excitation_TRI_max
            elif excitation_TRI < excitation_TRI_min:
                excitation_TRI = excitation_TRI_min
        if control_condition == 7: # 三角波
            if current_time < ramp_start:
                excitation_BIC = excitation_BIC_min
                excitation_TRI = excitation_TRI_min
            elif current_time <= switch_time+ramp_start:
                excitation_BIC = excitation_BIC_min  + (current_time - ramp_start) * (excitation_BIC_max - excitation_BIC_min) / (switch_time+ramp_start - ramp_start)
                excitation_TRI = excitation_TRI_min  + (current_time - ramp_start) * (excitation_TRI_max - excitation_TRI_min) / (switch_time+ramp_start - ramp_start)
            elif current_time <= switch_time*2+ramp_start:
                excitation_BIC = excitation_BIC_max + (-1) * (current_time - (switch_time+ramp_start)) * (excitation_BIC_max - excitation_BIC_min) / (switch_time*2+ramp_start - (switch_time+ramp_start))
                excitation_TRI = excitation_TRI_max + (-1) * (current_time - (switch_time+ramp_start)) * (excitation_TRI_max - excitation_TRI_min) / (switch_time*2+ramp_start - (switch_time+ramp_start))
            else:
                excitation_BIC = excitation_BIC_min
                excitation_TRI = excitation_TRI_min
        if control_condition == 8: # fuzzy制御
            if i % int(1/excitation_frequency/TIME_STEP) == 1:
                target_angle = target_angles[i]
                excitation_BIC += Fuzzy_controller(error_mfs, delta_v_values, rules, np.rad2deg(target_angle), np.rad2deg(ObservedAngle_rad[i-1]))
        excitation_BIC_input = excitation_BIC
        excitation_TRI_input = excitation_TRI

        # 遅れ時間要素追加時
        if delay_flag:
            excitation_BIC_input = delay_BIC[i%n_delay_samples_BIC]
            delay_BIC[i%n_delay_samples_BIC] = excitation_BIC
            excitation_TRI_input = delay_TRI[i%n_delay_samples_TRI]
            delay_TRI[i%n_delay_samples_TRI] = excitation_TRI
        
        # リクルートメント曲線使用時
        if recruitment_flag and recruitment_pattern != 8:
            if recruitment_pattern == 0:
                excitation_BIC_input = sigmoid_function(excitation_BIC_input, sh_REC, xc, excitation_BIC_min, excitation_BIC_max)
                excitation_TRI_input = sigmoid_function(excitation_TRI_input, sh_REC, xc, excitation_TRI_min, excitation_TRI_max)
            elif recruitment_pattern == 1:
                excitation_BIC_input = logarithmic_function(excitation_BIC_input, excitation_BIC_min, excitation_BIC_max)
                excitation_TRI_input = logarithmic_function(excitation_TRI_input, excitation_TRI_min, excitation_TRI_max)
            elif recruitment_pattern == 2:
                excitation_BIC_input = quadratic_function(excitation_BIC_input, excitation_BIC_min, excitation_BIC_max)
                excitation_TRI_input = quadratic_function(excitation_TRI_input, excitation_TRI_min, excitation_TRI_max)
            elif recruitment_pattern == 3:
                excitation_BIC_input = cubic_function(excitation_BIC_input, excitation_BIC_min, excitation_BIC_max)
                excitation_TRI_input = cubic_function(excitation_TRI_input, excitation_TRI_min, excitation_TRI_max)
            elif recruitment_pattern == 4:
                excitation_BIC_input = first_order_delay_system(excitation_BIC_input, excitation_BIC_min, excitation_BIC_max)
                excitation_TRI_input = first_order_delay_system(excitation_TRI_input, excitation_TRI_min, excitation_TRI_max)
            elif recruitment_pattern == 5:
                excitation_BIC_input = tanh(excitation_BIC_input, excitation_BIC_min, excitation_BIC_max)
                excitation_TRI_input = tanh(excitation_TRI_input, excitation_TRI_min, excitation_TRI_max)
            elif recruitment_pattern == 6:
                excitation_BIC_input = sqrt(excitation_BIC_input, excitation_BIC_min, excitation_BIC_max)
                excitation_TRI_input = sqrt(excitation_TRI_input, excitation_TRI_min, excitation_TRI_max)
            elif recruitment_pattern == 7:
                excitation_BIC_input = sine_squared(excitation_BIC_input, excitation_BIC_min, excitation_BIC_max)
                excitation_TRI_input = sine_squared(excitation_TRI_input, excitation_TRI_min, excitation_TRI_max)

        # 制御器に制御値をセット
        if target_muscle == 0 or target_muscle == 2:
            excitation_function_BIClong = osim.Constant(excitation_BIC_input)
            excitation_function_BICshort = osim.Constant(excitation_BIC_input)
            controller_BIClong.prescribeControlForActuator("BIClong", excitation_function_BIClong)
            controller_BICshort.prescribeControlForActuator("BICshort", excitation_function_BICshort)
            if target_muscle == 0:
                excitation_TRI_input = excitation_TRI_min
                excitation_TRI = excitation_TRI_min
        if target_muscle == 1 or target_muscle == 2:
            excitation_function_TRIlong = osim.Constant(excitation_TRI_input)
            excitation_function_TRIlat = osim.Constant(excitation_TRI_input)
            excitation_function_TRImed = osim.Constant(excitation_TRI_input)
            controller_TRIlong.prescribeControlForActuator("TRIlong", excitation_function_TRIlong)
            controller_TRIlat.prescribeControlForActuator("TRIlat", excitation_function_TRIlat)
            controller_TRImed.prescribeControlForActuator("TRImed", excitation_function_TRImed)
            if target_muscle == 1:
                excitation_BIC_input = excitation_BIC_min
                excitation_BIC = excitation_BIC_min

        # time_stepを一つ進めて積分実行
        state.setTime(current_time)
        state = manager.integrate(current_time + TIME_STEP)
        # データ保存
        Time[i] = current_time
        Excitation_BIC[i] = excitation_BIC
        Excitation_TRI[i] = excitation_TRI
        Excitation_BIC_input[i] = excitation_BIC_input
        Excitation_TRI_input[i] = excitation_TRI_input
        ObservedAngle_rad[i] = osimmodel.getStateVariableValues(state).get(2)
        ObservedAngularVelocity_rad[i] = osimmodel.getStateVariableValues(state).get(3)
        if TargetFlag:
            Error_rad[i] = target_angles[i] - osimmodel.getStateVariableValues(state).get(2)
        if control_condition == 6:
            Excitation_PID_BIC[i] = excitation_PID_BIC
            Excitation_ANN_BIC[i] = excitation_ANN_BIC
            TrainingData_Excitation_BIC[i] = trainingdata_excitation_BIC
            Excitation_PID_TRI[i] = excitation_PID_TRI
            Excitation_ANN_TRI[i] = excitation_ANN_TRI
            TrainingData_Excitation_TRI[i] = trainingdata_excitation_TRI
        if not iteration_flag:
            print("{}シミュレーション中:{}/{}".format(controller_name, round(current_time,2), end_time))
        elif iteration_flag:
            #print("{}回/{}回 {}シミュレーション中:{}/{}".format(iteration, total_iteration, controller_name, round(current_time,2), end_time))
            print("{}回/{}回 {}シミュレーション中:{}/{} {}".format(iteration, total_iteration, controller_name, round(current_time,2), end_time, np.rad2deg(ObservedAngle_rad[i])))

        Excitation_B[i] = BIClong.getExcitation(state)
        Activation_B[i] = BIClong.getActivation(state)
        FiberLength_B[i] = BIClong.getFiberLength(state)
        PennationAngle_B[i] = BIClong.getPennationAngle(state)
        CosPennationAngle_B[i] = BIClong.getCosPennationAngle(state)
        TendonLength_B[i] = BIClong.getTendonLength(state)
        NormalizedFiberLength_B[i] = BIClong.getNormalizedFiberLength(state)
        FiberLengthAlongTendon_B[i] = BIClong.getFiberLengthAlongTendon(state)
        TendonStrain_B[i] = BIClong.getTendonStrain(state)
        ActiveForceLengthMultiplier_B[i] = BIClong.getActiveForceLengthMultiplier(state)
        FiberVelocity_B[i] = BIClong.getFiberVelocity(state)
        NormalizedFiberVelocity_B[i] = BIClong.getNormalizedFiberVelocity(state)
        FiberVelocityAlongTendon_B[i] = BIClong.getFiberVelocityAlongTendon(state)
        TendonVelocity_B[i] = BIClong.getTendonVelocity(state)
        ForceVelocityMultiplier_B[i] = BIClong.getForceVelocityMultiplier(state)
        PennationAngularVelocity_B[i] = BIClong.getPennationAngularVelocity(state)
        FiberForce_B[i] = BIClong.getFiberForce(state)
        FiberForceAlongTendon_B[i] = BIClong.getFiberForceAlongTendon(state)
        ActiveFiberForce_B[i] = BIClong.getActiveFiberForce(state)
        PassiveFiberForce_B[i] = BIClong.getPassiveFiberForce(state)
        ActiveFiberForceAlongTendon_B[i] = BIClong.getActiveFiberForceAlongTendon(state)
        PassiveFiberForceAlongTendon_B[i] = BIClong.getPassiveFiberForceAlongTendon(state)
        TendonForce_B[i] = BIClong.getTendonForce(state)
        FiberStiffness_B[i] = BIClong.getFiberStiffness(state)
        FiberStiffnessAlongTendon_B[i] = BIClong.getFiberStiffnessAlongTendon(state)
        TendonStiffness_B[i] = BIClong.getTendonStiffness(state)
        MuscleStiffness_B[i] = BIClong.getMuscleStiffness(state)
        FiberActivePower_B[i] = BIClong.getFiberActivePower(state)
        FiberPassivePower_B[i] = BIClong.getFiberPassivePower(state)
        TendonPower_B[i] = BIClong.getTendonPower(state)
        MusclePower_B[i] = BIClong.getMusclePower(state)
        FiberPotentialEnergy_B[i] = BIClong.getFiberPotentialEnergy(state)
        TendonPotentialEnergy_B[i] = BIClong.getTendonPotentialEnergy(state)
        MusclePotentialEnergy_B[i] = BIClong.getMusclePotentialEnergy(state)
        PassiveForceMultiplier_B[i] = BIClong.getPassiveForceMultiplier(state)

        Excitation_T[i] = TRIlong.getExcitation(state)
        Activation_T[i] = TRIlong.getActivation(state)
        FiberLength_T[i] = TRIlong.getFiberLength(state)
        PennationAngle_T[i] = TRIlong.getPennationAngle(state)
        CosPennationAngle_T[i] = TRIlong.getCosPennationAngle(state)
        TendonLength_T[i] = TRIlong.getTendonLength(state)
        NormalizedFiberLength_T[i] = TRIlong.getNormalizedFiberLength(state)
        FiberLengthAlongTendon_T[i] = TRIlong.getFiberLengthAlongTendon(state)
        TendonStrain_T[i] = TRIlong.getTendonStrain(state)
        ActiveForceLengthMultiplier_T[i] = TRIlong.getActiveForceLengthMultiplier(state)
        FiberVelocity_T[i] = TRIlong.getFiberVelocity(state)
        NormalizedFiberVelocity_T[i] = TRIlong.getNormalizedFiberVelocity(state)
        FiberVelocityAlongTendon_T[i] = TRIlong.getFiberVelocityAlongTendon(state)
        TendonVelocity_T[i] = TRIlong.getTendonVelocity(state)
        ForceVelocityMultiplier_T[i] = TRIlong.getForceVelocityMultiplier(state)
        PennationAngularVelocity_T[i] = TRIlong.getPennationAngularVelocity(state)
        FiberForce_T[i] = TRIlong.getFiberForce(state)
        FiberForceAlongTendon_T[i] = TRIlong.getFiberForceAlongTendon(state)
        ActiveFiberForce_T[i] = TRIlong.getActiveFiberForce(state)
        PassiveFiberForce_T[i] = TRIlong.getPassiveFiberForce(state)
        ActiveFiberForceAlongTendon_T[i] = TRIlong.getActiveFiberForceAlongTendon(state)
        PassiveFiberForceAlongTendon_T[i] = TRIlong.getPassiveFiberForceAlongTendon(state)
        TendonForce_T[i] = TRIlong.getTendonForce(state)
        FiberStiffness_T[i] = TRIlong.getFiberStiffness(state)
        FiberStiffnessAlongTendon_T[i] = TRIlong.getFiberStiffnessAlongTendon(state)
        TendonStiffness_T[i] = TRIlong.getTendonStiffness(state)
        MuscleStiffness_T[i] = TRIlong.getMuscleStiffness(state)
        FiberActivePower_T[i] = TRIlong.getFiberActivePower(state)
        FiberPassivePower_T[i] = TRIlong.getFiberPassivePower(state)
        TendonPower_T[i] = TRIlong.getTendonPower(state)
        MusclePower_T[i] = TRIlong.getMusclePower(state)
        FiberPotentialEnergy_T[i] = TRIlong.getFiberPotentialEnergy(state)
        TendonPotentialEnergy_T[i] = TRIlong.getTendonPotentialEnergy(state)
        MusclePotentialEnergy_T[i] = TRIlong.getMusclePotentialEnergy(state)
        PassiveForceMultiplier_T[i] = TRIlong.getPassiveForceMultiplier(state)

        # グラフ描画用
        if plot_flag:
            current_times.append(current_time)
            state_values1.append(np.degrees(ObservedAngle_rad[i]))
            state_values2.append(Excitation_BIC[i])
            state_values3.append(Excitation_TRI[i])
            if TargetFlag:
                state_values4.append(np.degrees(target_angles[i]))
            if i % int(0.5/TIME_STEP) == 0:
                ax1.cla()  # グラフクリア
                ax1.plot(current_times, state_values1, label='Observed Angle')  # 新しいデータでプロット
                if TargetFlag:
                    ax1.plot(current_times, state_values4, label='Target Angle')  # 新しいデータでプロット
                ax1.set_xlabel('Time [sec]')
                ax1.set_ylabel('Elbow Joint Angle [deg]')
                ax1.legend()
                ax2.cla()  # グラフクリア
                ax2.plot(current_times, state_values2, label='Excitation_BIC')  # 新しいデータでプロット
                ax2.plot(current_times, state_values3, label='Excitation_TRI')  # 新しいデータでプロット
                ax2.set_xlabel('Time [sec]')
                ax2.set_ylabel('Excitation Intensity')
                ax2.legend()
                plt.draw()
                plt.pause(0.01)
    # シミュレーション終了時の動作
    TIME_SIMULATION_END = time.time()
    elapsed_time = TIME_SIMULATION_END - TIME_SIMULATION_START
    print("シミュレーション終了")
    print(f"シミュレーション時間: {elapsed_time} 秒")

    # データ格納
    ObservedAngle_degree = np.degrees(ObservedAngle_rad)
    ObservedAngularVelocity_degree = np.degrees(ObservedAngularVelocity_rad)
    ObservedAngularAcceleration_degree[0] = 0
    for i in range(n_samples): # 角加速度
        if i == 0:
            ObservedAngularAcceleration_degree[i] = 0
        else:
            ObservedAngularAcceleration_degree[i] = (ObservedAngularVelocity_degree[i] - ObservedAngularVelocity_degree[i-1]) / TIME_STEP
    if TargetFlag:
        for i in range(n_samples):
            TargetAngle_degree[i] = np.degrees(target_angles[i])
            TargetAngularVelocity_degree[i] = np.degrees(target_velocities[i])
            TargetAngularAcceleration_degree[i] = np.degrees(target_accelerations[i])
        Error_degree = np.degrees(Error_rad)
        Mae = np.average(np.abs(Error_degree[for_sec_index:]))
        Mae_col[0] = Mae
        Rmse = np.sqrt(np.average(np.square(Error_degree[for_sec_index:])))
        print(Rmse)
        Rmse_col[0] = Rmse
        Est_Lag_col[0] = calculate_estimation_lag(TargetAngle_degree[for_sec_index:], ObservedAngle_degree[for_sec_index:], TIME_STEP)
    end_time_col[0] = end_time
    time_step_col[0] = TIME_STEP
    elapsed_time_col[0] = elapsed_time
    T_col[0] = T1
    T_col[1] = T2
    T_col[2] = T3
    Excitation_BIC[0] = Excitation_BIC[1]
    Excitation_ANN_BIC[0] = Excitation_ANN_BIC[1]
    TrainingData_Excitation_BIC[0] = TrainingData_Excitation_BIC[1]
    Excitation_TRI[0] = Excitation_TRI[1]
    Excitation_ANN_TRI[0] = Excitation_ANN_TRI[1]
    TrainingData_Excitation_TRI[0] = TrainingData_Excitation_TRI[1]
    Excitation_BIC_norm = normalize_list(Excitation_BIC, 0, excitation_BIC_max)
    Excitation_PID_BIC_norm = normalize_list(Excitation_PID_BIC, 0, excitation_BIC_max)
    Excitation_ANN_BIC_norm = normalize_list(Excitation_ANN_BIC, 0, excitation_BIC_max)
    TrainingData_Excitation_BIC_norm = normalize_list(TrainingData_Excitation_BIC, 0, excitation_BIC_max)
    Excitation_TRI_norm = normalize_list(Excitation_TRI, 0, excitation_TRI_max)
    Excitation_PID_TRI_norm = normalize_list(Excitation_PID_TRI, 0, excitation_TRI_max)
    Excitation_ANN_TRI_norm = normalize_list(Excitation_ANN_TRI, 0, excitation_TRI_max)
    TrainingData_Excitation_TRI_norm = normalize_list(TrainingData_Excitation_TRI, 0, excitation_TRI_max)


    if max(ObservedAngle_degree) > 180 or min(ObservedAngle_degree) < 0:
        bug_flag = True

    if control_condition == 6:
        for i in range(len(Excitation_PID_BIC) - for_sec_index):
            Power_Ratio_ANN_BIC += 100 * (Excitation_ANN_BIC[i + for_sec_index] ** 2) / (Excitation_ANN_BIC[i + for_sec_index] ** 2 + Excitation_PID_BIC[i + for_sec_index] ** 2)
            Power_Ratio_ANN_TRI += 100 * (Excitation_ANN_TRI[i + for_sec_index] ** 2) / (Excitation_ANN_TRI[i + for_sec_index] ** 2 + Excitation_PID_TRI[i + for_sec_index] ** 2)
        Power_Ratio_ANN_BIC /= (len(Excitation_PID_BIC) - for_sec_index)
        Power_Ratio_ANN_TRI /= (len(Excitation_PID_BIC) - for_sec_index)
        Power_Ratio_ANN_BIC_col[0] = Power_Ratio_ANN_BIC
        Power_Ratio_ANN_TRI_col[0] = Power_Ratio_ANN_TRI
        if iteration >= 11 and (iteration - 1) % 10 == 0:
            if (np.average(training_history[5:10]) - np.average(training_history[0:5])) < 0.5:
                training_history[10] += 1
            if (np.average(training_history[16:21]) - np.average(training_history[11:16])) < 0.5:
                training_history[21] += 1
        training_history[(iteration - 1) % 10] = Power_Ratio_ANN_BIC
        training_history[(iteration - 1) % 10 + 11] = Power_Ratio_ANN_TRI

    # 以下、csvへシミュレーション結果の保存
    print("結果保存中")
    data = {
        "Time": Time,
        "Excitation_BIC": Excitation_BIC,
        "Excitation_PID_BIC": Excitation_PID_BIC,
        "Excitation_ANN_BIC": Excitation_ANN_BIC,
        "TrainingData_Excitation_BIC": TrainingData_Excitation_BIC,
        "Excitation_TRI": Excitation_TRI,
        "Excitation_PID_TRI": Excitation_PID_TRI,
        "Excitation_ANN_TRI": Excitation_ANN_TRI,
        "TrainingData_Excitation_TRI": TrainingData_Excitation_TRI,
        "Excitation_BIC_norm": Excitation_BIC_norm,
        "Excitation_PID_BIC_norm": Excitation_PID_BIC_norm,
        "Excitation_ANN_BIC_norm": Excitation_ANN_BIC_norm,
        "TrainingData_Excitation_BIC_norm": TrainingData_Excitation_BIC_norm,
        "Excitation_TRI_norm": Excitation_TRI_norm,
        "Excitation_PID_TRI_norm": Excitation_PID_TRI_norm,
        "Excitation_ANN_TRI_norm": Excitation_ANN_TRI_norm,
        "TrainingData_Excitation_TRI_norm": TrainingData_Excitation_TRI_norm,
        "ObservedAngle": ObservedAngle_degree,
        "TargetAngle": TargetAngle_degree,
        "ErrorAngle": Error_degree,
        "ObservedAngularVelocity": ObservedAngularVelocity_degree,
        "TargetAngularVelocity": TargetAngularVelocity_degree,
        "ObservedAngularAcceleration": ObservedAngularAcceleration_degree,
        "TargetAngularAcceleration": TargetAngularAcceleration_degree,
        "end_time": end_time_col,
        "TIME_STEP": time_step_col,
        "elapsed_time": elapsed_time_col,
        "Power_Ratio_ANN_BIC": Power_Ratio_ANN_BIC_col,
        "Power_Ratio_ANN_TRI": Power_Ratio_ANN_TRI_col,
        "MAE": Mae_col,
        "RMSE": Rmse_col,
        "Est_Lag": Est_Lag_col,
        "T": T_col,
        "cost": costs_col,
        "f_x": f_x_col,
        "Kp": kp_col,
        "Ki": ki_col,
        "Kp_hat": kp_hat_col,
        "Ki_hat": ki_hat_col,
        "Kd_hat": kd_hat_col,
        "noise_Kp": noise_kp_col,
        "noise_Ki": noise_ki_col,
        "noise_Kd": noise_kd_col,
        "Excitation_BIC_input": Excitation_BIC_input,
        "Excitation_TRI_input": Excitation_TRI_input,
        "Kd": kd_col,
        "Excitation_B":Excitation_B,
        "Activation_B":Activation_B,
        "FiberLength_B":FiberLength_B,
        "PennationAngle_B":PennationAngle_B,
        "CosPennationAngle_B":CosPennationAngle_B,
        "TendonLength_B":TendonLength_B,
        "NormalizedFiberLength_B":NormalizedFiberLength_B,
        "FiberLengthAlongTendon_B":FiberLengthAlongTendon_B,
        "TendonStrain_B":TendonStrain_B,
        "PassiveForceMultiplier_B":PassiveForceMultiplier_B,
        "ActiveForceLengthMultiplier_B":ActiveForceLengthMultiplier_B,
        "FiberVelocity_B":FiberVelocity_B,
        "NormalizedFiberVelocity_B":NormalizedFiberVelocity_B,
        "FiberVelocityAlongTendon_B":FiberVelocityAlongTendon_B,
        "TendonVelocity_B":TendonVelocity_B,
        "ForceVelocityMultiplier_B":ForceVelocityMultiplier_B,
        "PennationAngularVelocity_B":PennationAngularVelocity_B,
        "FiberForce_B":FiberForce_B,
        "FiberForceAlongTendon_B":FiberForceAlongTendon_B,
        "ActiveFiberForce_B":ActiveFiberForce_B,
        "PassiveFiberForce_B":PassiveFiberForce_B,
        "ActiveFiberForceAlongTendon_B":ActiveFiberForceAlongTendon_B,
        "PassiveFiberForceAlongTendon_B":PassiveFiberForceAlongTendon_B,
        "TendonForce_B":TendonForce_B,
        "FiberStiffness_B":FiberStiffness_B,
        "FiberStiffnessAlongTendon_B":FiberStiffnessAlongTendon_B,
        "TendonStiffness_B":TendonStiffness_B,
        "MuscleStiffness_B":MuscleStiffness_B,
        "FiberActivePower_B":FiberActivePower_B,
        "FiberPassivePower_B":FiberPassivePower_B,
        "TendonPower_B":TendonPower_B,
        "MusclePower_B":MusclePower_B,
        "Stress_B":Stress_B,
        "FiberPotentialEnergy_B":FiberPotentialEnergy_B,
        "TendonPotentialEnergy_B":TendonPotentialEnergy_B,
        "MusclePotentialEnergy_B":MusclePotentialEnergy_B,
        "Excitation_T":Excitation_T,
        "Activation_T":Activation_T,
        "FiberLength_T":FiberLength_T,
        "PennationAngle_T":PennationAngle_T,
        "CosPennationAngle_T":CosPennationAngle_T,
        "TendonLength_T":TendonLength_T,
        "NormalizedFiberLength_T":NormalizedFiberLength_T,
        "FiberLengthAlongTendon_T":FiberLengthAlongTendon_T,
        "TendonStrain_T":TendonStrain_T,
        "PassiveForceMultiplier_T":PassiveForceMultiplier_T,
        "ActiveForceLengthMultiplier_T":ActiveForceLengthMultiplier_T,
        "FiberVelocity_T":FiberVelocity_T,
        "NormalizedFiberVelocity_T":NormalizedFiberVelocity_T,
        "FiberVelocityAlongTendon_T":FiberVelocityAlongTendon_T,
        "TendonVelocity_T":TendonVelocity_T,
        "ForceVelocityMultiplier_T":ForceVelocityMultiplier_T,
        "PennationAngularVelocity_T":PennationAngularVelocity_T,
        "FiberForce_T":FiberForce_T,
        "FiberForceAlongTendon_T":FiberForceAlongTendon_T,
        "ActiveFiberForce_T":ActiveFiberForce_T,
        "PassiveFiberForce_T":PassiveFiberForce_T,
        "ActiveFiberForceAlongTendon_T":ActiveFiberForceAlongTendon_T,
        "PassiveFiberForceAlongTendon_T":PassiveFiberForceAlongTendon_T,
        "TendonForce_T":TendonForce_T,
        "FiberStiffness_T":FiberStiffness_T,
        "FiberStiffnessAlongTendon_T":FiberStiffnessAlongTendon_T,
        "TendonStiffness_T":TendonStiffness_T,
        "MuscleStiffness_T":MuscleStiffness_T,
        "FiberActivePower_T":FiberActivePower_T,
        "FiberPassivePower_T":FiberPassivePower_T,
        "TendonPower_T":TendonPower_T,
        "MusclePower_T":MusclePower_T,
        "Stress_T":Stress_T,
        "FiberPotentialEnergy_T":FiberPotentialEnergy_T,
        "TendonPotentialEnergy_T":TendonPotentialEnergy_T,
        "MusclePotentialEnergy_T":MusclePotentialEnergy_T,
}
    df = pd.DataFrame(data)
    # 現在のディレクトリと新しいフォルダ名を結合して新しいパスを作成
    folder_path = os.path.join(os.getcwd(), result_folder_name)
    # フォルダが存在しない場合は作成
    if not os.path.exists(folder_path):
        os.makedirs(folder_path)
    # Excelファイルの保存パス
    if not iteration_flag:
        csv_filename = os.path.join(folder_path, '{}_{}.csv'.format(controller_name, datetime.datetime.now().strftime("%m-%d_%H-%M-%S")))
    else:
        csv_filename = os.path.join(folder_path, '{}_{}_{}.csv'.format(iteration-1, controller_name, datetime.datetime.now().strftime("%m-%d_%H-%M-%S")))
    if bug_flag:
        csv_filename = os.path.join(folder_path, 'バグ{}_{}_{}.csv'.format(iteration-1, controller_name, datetime.datetime.now().strftime("%m-%d_%H-%M-%S")))
    # csvにデータ保存
    df.to_csv(csv_filename, index=False)
    print("保存完了")

    if control_condition == 6 and not bug_flag:
        # Excelファイルへのパスを定義
        result_dir = os.path.join(os.getcwd(), result_folder_name)
        if not os.path.exists(result_dir):
            os.makedirs(result_dir)
        file_path = os.path.join(result_dir, 'history.xlsx')
        
        # Excelファイルに保存するデータを準備
        data = {
            'Iteration': [iteration-1],
            'Power_Ratio_ANN_BIC': [Power_Ratio_ANN_BIC],
            'Power_Ratio_ANN_TRI': [Power_Ratio_ANN_TRI],
            'MAE': [Mae],
            'RMSE': [Rmse],
            'lr_BIC': [training_history[10]],
            'lr_TRi': [training_history[21]],
            "Est lag": [Est_Lag_col[0]]
        }
        df_new = pd.DataFrame(data)
        
        # ファイルが既に存在するか確認し、適切に処理
        if os.path.exists(file_path):
            # ファイルが存在する場合、読み込んで新しいデータを追加
            df_existing = pd.read_excel(file_path)
            df_updated = pd.concat([df_existing, df_new], ignore_index=True)
        else:
            # ファイルが存在しない場合、新しいファイルを作成
            df_updated = df_new
        
        # Excelファイルにデータを保存
        with pd.ExcelWriter(file_path, engine='openpyxl', mode='w') as writer:
            df_updated.to_excel(writer, index=False)

    # pre_learned_model.kerasの再学習
    if ANN_training_flag:
        if control_condition == 6 and not bug_flag:
            # 出力データの正規化
            custom_scaler_outputs_BIC = CustomScaler(-0.1, excitation_max_minus_min_BIC)
            custom_scaler_outputs_TRI = CustomScaler(-0.1, excitation_max_minus_min_TRI)
            outputs_BIC = custom_scaler_outputs_BIC.fit_transform(TrainingData_Excitation_BIC)
            outputs_TRI = custom_scaler_outputs_TRI.fit_transform(TrainingData_Excitation_TRI)
            outputs = np.vstack([outputs_BIC[for_sec_index:], outputs_TRI[for_sec_index:]]).T
            inputs = inputs[for_sec_index:]
            # モデルのコンパイル
            lr = 0.001 / (2 ** training_history[10])
            batch_size_num = int(0.5 / TIME_STEP)
            ANNmodel.compile(optimizer=keras.optimizers.Adam(learning_rate=0.001), loss='mean_squared_error')
            # モデルの学習
            ANNmodel.fit(inputs, outputs, epochs=1, batch_size=batch_size_num)
            # モデルの保存
            model_save_path = f"{model_folder_name}/{ANN_forder_name}/pre_learned_model_{iteration}.keras"
            ANNmodel.save(model_save_path)
            print(f"Model saved to {model_save_path}")

    if plot_flag:
        plt.ioff()
        plt.show(block=False)  # グラフを非ブロッキングモードで表示
        time.sleep(0.5)  # 5秒間待機
        plt.close()  # グラフウィンドウを閉じる

    return bug_flag, training_history