import sys
import numpy as np
import pandas as pd
import os
import time
import datetime
import tensorflow as tf
from tensorflow import keras
import random
import matplotlib.pyplot as plt
from sklearn.model_selection import train_test_split
import opensim as osim

current_dir = os.path.dirname(os.path.abspath(__file__))
parent_dir = os.path.abspath(os.path.join(current_dir, '..', '..'))
sys.path.append(parent_dir)
from program.sub.set_initial_state_from_sto import set_initial_state_from_sto
from program.sub.Est_Lag import calculate_estimation_lag

def control_main(control_condition, target_muscle, target_pattern, quiet_time, end_time, excitation_BIC_max, excitation_TRI_max, zero_point, amplitude, frequency, model_file_name, model_folder_name, IniSta_file_name, IniSta_folder_name, result_folder_name, iteration_flag, iteration, total_iteration, training_history, TIME_STEP):
    # BICのPID制御器
    def PID_BIC(target_angle, ThisAngle, LastError_BIC, LastIntegralError_BIC, excitation_ANN_BIC):
        PID_paras_default = [0.28658910483018335, 0.48312854373915975, 0.011770462075924547]
        PID_paras_type1_before_tuning = [0.21725388137889728, 0.3068688319798013, 0.01861378741795764]
        PID_paras_type1_after_tuning = [0.4, 0.32, 0.055]
        PID_paras_powerful = [0.23008799710480657, 0.5110710649869753, 0.007738250474873482]
        PID_paras_athletic = [0.23856302849874864, 0.48860902272830575, 0.008516428452856157]
        PID_paras_thin = [0.31250825396940446, 0.5810201526494893, 0.013529695302399782]
        PID_paras_sedentary = [0.35577796573236514, 0.6107479370192849, 0.016545730703365735]
        PID_paras_elderly = [0.38328811162179227, 0.6466250844230451, 0.018076548911134055]
        PID_paras_fat = [0.37760918533851484, 0.6550972459000699, 0.017805342219391705]
        PID_paras_weak = [0.46613926128577005, 0.7519584682159091, 0.023532960141594988]

        Kp_BIC = PID_paras_weak[0]
        Ki_BIC = PID_paras_weak[1]
        Kd_BIC = PID_paras_weak[2]
        
        ThisError_BIC = target_angle - ThisAngle
        if not excitation_frequency_flag:
            ThisIntegralError_BIC = LastIntegralError_BIC + ThisError_BIC * TIME_STEP
            excitation_PID_BIC = Kp_BIC * ThisError_BIC + Kd_BIC * (ThisError_BIC - LastError_BIC) / TIME_STEP + Ki_BIC * ThisIntegralError_BIC
        else:
            ThisIntegralError_BIC = LastIntegralError_BIC + ThisError_BIC * (1 / excitation_frequency)
            excitation_PID_BIC = Kp_BIC * ThisError_BIC + Kd_BIC * (ThisError_BIC - LastError_BIC) / (1 / excitation_frequency) + Ki_BIC * ThisIntegralError_BIC
        # Anti Reset Windup
        if excitation_PID_BIC + excitation_ANN_BIC > excitation_max_min_BIC:
            ThisIntegralError_BIC = LastIntegralError_BIC
        elif excitation_PID_BIC + excitation_ANN_BIC < 0:
            ThisIntegralError_BIC = LastIntegralError_BIC
        return excitation_PID_BIC, ThisError_BIC, ThisIntegralError_BIC
    # TRIのPID制御器
    def PID_TRI(target_angle, ThisAngle, LastError_TRI, LastIntegralError_TRI, excitation_ANN_TRI):
        Kp_TRI = -0.451607540911566
        Ki_TRI = -2.0767583542786556
        Kd_TRI = -0.011265404393685494
        ThisError_TRI = target_angle - ThisAngle
        if not excitation_frequency_flag:
            ThisIntegralError_TRI = LastIntegralError_TRI + ThisError_TRI * TIME_STEP
            excitation_PID_TRI = Kp_TRI * ThisError_TRI + Kd_TRI * (ThisError_TRI - LastError_TRI) / TIME_STEP + Ki_TRI * ThisIntegralError_TRI
        else:
            ThisIntegralError_TRI = LastIntegralError_TRI + ThisError_TRI * (1 / excitation_frequency)
            excitation_PID_TRI = Kp_TRI * ThisError_TRI + Kd_TRI * (ThisError_TRI - LastError_TRI) / (1 / excitation_frequency) + Ki_TRI * ThisIntegralError_TRI
        # Anti Reset Windup
        if excitation_PID_TRI + excitation_ANN_TRI > excitation_max_min_TRI:
            ThisIntegralError_TRI = LastIntegralError_TRI
        elif excitation_PID_TRI + excitation_ANN_TRI < 0:
            ThisIntegralError_TRI = LastIntegralError_TRI
        return excitation_PID_TRI, ThisError_TRI, ThisIntegralError_TRI

    # 目標軌道
    def Target_Angle(target_pattern, time_array):
        target_angle = np.zeros_like(time_array)
        target_velocity = np.zeros_like(time_array)
        target_acceleration = np.zeros_like(time_array)
        if target_pattern == 0:  # 矩形波_周期2秒
            for i, t in enumerate(time_array):
                if t % 4 < 2:
                    target_angle[i] = np.radians(40)
                else:
                    target_angle[i] = np.radians(80)
        elif target_pattern == 1:  # 矩形波_周期4秒
            for i, t in enumerate(time_array):
                if t % 8 < 4:
                    target_angle[i] = np.radians(40)
                else:
                    target_angle[i] = np.radians(80)
        elif target_pattern == 2:  # 矩形波_周期6秒
            for i, t in enumerate(time_array):
                if t % 12 < 6:
                    target_angle[i] = np.radians(40)
                else:
                    target_angle[i] = np.radians(80)
        elif target_pattern == 4:  # 正弦波
            A = 30  # 振幅
            f = 0.2  # 周波数 Hz
            target_angle = np.radians(A * np.sin(2 * PI * f * time_array) + 50)
        elif target_pattern == 5:  # ランダム波形
            global T1, T2, T3
            T1 = random.uniform(1.8, 3.2)
            T2 = random.uniform(3.2, 4.6)
            T3 = random.uniform(4.6, 6.0)
            span = 1
            if span == 0: # 小
                T1 = 1.8
                T2 = 3.2
                T3 = 4.6
            if span == 1: # 中
                T1 = 2.5
                T2 = 3.9
                T3 = 5.3
            if span == 2: # 大
                T1 = 3.2
                T2 = 4.6
                T3 = 6.0
            f1 = 1/T1
            f2 = 1/T2
            f3 = 1/T3
            y_angle_1 = np.sin(2 * PI * f1 * time_array)
            y_angle_2 = np.sin(2 * PI * f2 * time_array)
            y_angle_3 = np.sin(2 * PI * f3 * time_array)
            y_angle = y_angle_1 + y_angle_2 + y_angle_3
            y_velocity_1 = 2 * PI * f1 * np.cos(2 * PI * f1 * time_array)
            y_velocity_2 = 2 * PI * f2 * np.cos(2 * PI * f2 * time_array)
            y_velocity_3 = 2 * PI * f3 * np.cos(2 * PI * f3 * time_array)
            y_velocity = y_velocity_1 + y_velocity_2 + y_velocity_3
            y_accel_1 = -((2 * PI * f1)**2) * np.sin(2 * PI * f1 * time_array)
            y_accel_2 = -((2 * PI * f2)**2) * np.sin(2 * PI * f2 * time_array)
            y_accel_3 = -((2 * PI * f3)**2) * np.sin(2 * PI * f3 * time_array)
            y_accel = y_accel_1 + y_accel_2 + y_accel_3
            for i, t in enumerate(time_array):
                if t <= 1:
                    target_angle[i] = INITIAL_ANGLE_RAD
                    target_velocity[i] = 0
                    target_acceleration[i] = 0
                elif t <= 2:
                    target_angle[i] = INITIAL_ANGLE_RAD + (np.radians(60) - INITIAL_ANGLE_RAD) * (t - 1)
                    target_velocity[i] = np.radians(60) - INITIAL_ANGLE_RAD
                    target_acceleration[i] = 0
                elif t <= 4:
                    target_angle[i] = INITIAL_ANGLE_RAD + (np.radians(60) - INITIAL_ANGLE_RAD)
                    target_velocity[i] = 0
                    target_acceleration[i] = 0
                else:
                    index = i - int(4 / TIME_STEP)
                    target_angle[i] = INITIAL_ANGLE_RAD + (np.radians(60) - INITIAL_ANGLE_RAD) + np.radians(10) * y_angle[index]
                    target_velocity[i] = np.radians(10) * y_velocity[index]
                    target_acceleration[i] = np.radians(10) * y_accel[index]
        elif target_pattern == 6:  # Step
            for i, t in enumerate(time_array):
                if t <= 1:
                    target_angle[i] = INITIAL_ANGLE_RAD
                elif t <= end_time:
                    target_angle[i] = np.radians(80)
        return target_angle, target_velocity, target_acceleration
    
    class CustomScaler:
        def __init__(self, min_val, max_val):
            self.min_val = min_val
            self.max_val = max_val
            
        def fit(self, X):
            return self
        
        def transform(self, X):
            scaled = (X - self.min_val) / (self.max_val - self.min_val)
            return scaled
        
        def fit_transform(self, X):
            self.fit(X)
            return self.transform(X)
        
    class CustomInverseScaler:
        def __init__(self, min_val, max_val):
            self.min_val = min_val
            self.max_val = max_val
            self.data_range = max_val - min_val

        def fit(self, X_scaled):
            return self
        
        def transform(self, X_scaled):
            X_original = X_scaled * self.data_range + self.min_val
            return X_original
            
        def inverse_transform(self, X_scaled):
            self.fit(X_scaled)
            return self.transform(X_scaled)
    
    def get_text(file_path, target_line_num, start_pos, end_pos):
        # ファイルを開く
        with open(file_path, "r", encoding="utf-8") as file:
            # ファイル内の全行をリストとして取得する
            lines = file.readlines()
            # 指定した行番号が範囲内であるかを確認
            if 0 <= target_line_num - 1 < len(lines):
                # 指定した行の部分文字列をスライス
                target_line = lines[target_line_num - 1]
                extracted_text = target_line[start_pos:end_pos]
            else:
                print(f"行 {target_line_num} はファイルの範囲外です。")
                exit()
        return extracted_text
        
    # 諸変数
    start_time = 0.0
    ramp_start = quiet_time # Ramp開始
    ramp_end = end_time # Ramp終了
    step_start = quiet_time # Step開始
    step_end = end_time # Step終了
    ramp2step = 1 + quiet_time # Ramp→Stepの切り替え時間(Ramp+Step)
    switch_time = (end_time / 2) - quiet_time
    sin_start = 7
    EXCITATION_MIN  = 0.01
    excitation_BIC = EXCITATION_MIN
    excitation_TRI = EXCITATION_MIN
    excitation_max_min_BIC = excitation_BIC_max - EXCITATION_MIN
    excitation_max_min_TRI = excitation_TRI_max - EXCITATION_MIN
    excitation_PID_BIC = 0
    excitation_ANN_BIC = 0
    Power_Ratio_ANN_BIC = 0
    Power_Ratio_ANN_TRI = 0
    PI = np.pi
    initial_sto_file = "{}/{}".format(IniSta_folder_name, IniSta_file_name)
    INITIAL_ANGLE_RAD = np.radians(float(get_text(initial_sto_file, 8, 31, 44)))
    global T1, T2, T3
    T1 = 0
    T2 = 0
    T3 = 0
    for_sec_index = int(4 / TIME_STEP)  # 3秒に相当するインデックス

    # flag
    bug_flag = False
    plot_flag = False
    ANN_training_flag = True
    excitation_frequency_flag = False
    square_wave_flag = False
    excitation_frequency = 20
    downsamplerate = int(1/excitation_frequency/TIME_STEP)
    pulse_width = 300 / (10**6)
    if square_wave_flag:
        time_array_square_wave = np.arange(0, end_time + TIME_STEP, TIME_STEP)
        square_wave = np.zeros_like(time_array_square_wave)
        for i, t in enumerate(time_array_square_wave):
            if (t+TIME_STEP) % (1/excitation_frequency) < pulse_width:
                square_wave[i] = 1

    # 制御器の名前
    if control_condition == 0:
        controller_name = "Step"
        TargetFlag = False
    elif control_condition == 1:
        controller_name = "Ramp"
        TargetFlag = False
    elif control_condition == 2:
        controller_name = "PID"
        TargetFlag = True
    elif control_condition == 3:
        controller_name = "Ramp2Step"
        TargetFlag = False
    elif control_condition == 5:
        controller_name = "Sin_Wave"
        TargetFlag = False
    elif control_condition == 6:
        controller_name = "FEL"
        TargetFlag = True
    elif control_condition == 7:
        controller_name = "Triangle_Wage"
        TargetFlag = False

    # モデル初期設定
    osimmodel = osim.Model("{}/{}".format(model_folder_name, model_file_name))
    state = osimmodel.initSystem()
    state = set_initial_state_from_sto(osimmodel, initial_sto_file)
    BIClong = osimmodel.getMuscles().get("BIClong")
    BICshort = osimmodel.getMuscles().get("BICshort")
    TRIlong = osimmodel.getMuscles().get("TRIlong")
    TRIlat = osimmodel.getMuscles().get("TRIlat")
    TRImed = osimmodel.getMuscles().get("TRImed")
    Bra = osimmodel.getMuscles().get("BRA")

    # パラメータ格納用変数
    n_samples = int(end_time/TIME_STEP) + 1
    n_samples_down = int(end_time/TIME_STEP/downsamplerate) + 1
    Time = np.full(n_samples, np.nan)
    Time_down = np.full(n_samples, np.nan)
    Excitation_BIC = np.full(n_samples, np.nan)
    Excitation_BIC_down = np.full(n_samples, np.nan)
    Excitation_ANN_BIC = np.full(n_samples, np.nan)
    Excitation_ANN_BIC_down = np.full(n_samples, np.nan)
    Excitation_PID_BIC = np.full(n_samples, np.nan)
    Excitation_PID_BIC_down = np.full(n_samples, np.nan)
    TrainingData_Excitation_BIC = np.full(n_samples, np.nan)
    TrainingData_Excitation_BIC_down = np.full(n_samples, np.nan)
    Excitation_TRI = np.full(n_samples, np.nan)
    Excitation_TRI_down = np.full(n_samples, np.nan)
    Excitation_ANN_TRI = np.full(n_samples, np.nan)
    Excitation_ANN_TRI_down = np.full(n_samples, np.nan)
    Excitation_PID_TRI = np.full(n_samples, np.nan)
    Excitation_PID_TRI_down = np.full(n_samples, np.nan)
    TrainingData_Excitation_TRI = np.full(n_samples, np.nan)
    TrainingData_Excitation_TRI_down = np.full(n_samples, np.nan)
    Activation_BIC = np.full(n_samples, np.nan)
    Activation_down = np.full(n_samples, np.nan)
    ObservedAngle_rad = np.full(n_samples, np.nan)
    ObservedAngle_rad_down = np.full(n_samples, np.nan)
    ObservedAngle_degree = np.full(n_samples, np.nan)
    ObservedAngle_degree_down = np.full(n_samples, np.nan)
    TargetAngle_degree = np.full(n_samples, np.nan)
    TargetAngle_degree_down = np.full(n_samples, np.nan)
    ObservedAngularVelocity_rad = np.full(n_samples, np.nan)
    ObservedAngularVelocity_degree = np.full(n_samples, np.nan)
    ObservedAngularVelocity_degree_down = np.full(n_samples, np.nan)
    TargetAngularVelocity_degree = np.full(n_samples, np.nan)
    TargetAngularVelocity_degree_down = np.full(n_samples, np.nan)
    ObservedAngularAcceleration_degree = np.full(n_samples, np.nan)
    ObservedAngularAcceleration_degree_down = np.full(n_samples, np.nan)
    TargetAngularAcceleration_degree = np.full(n_samples, np.nan)
    TargetAngularAcceleration_degree_down = np.full(n_samples, np.nan)
    Error_rad = np.full(n_samples, np.nan)
    Error_rad_down = np.full(n_samples, np.nan)
    Error_degree = np.full(n_samples, np.nan)
    Error_degree_down = np.full(n_samples, np.nan)
    Mae_col = np.full(n_samples, np.nan)
    Mae_col_down = np.full(n_samples, np.nan)
    Rmse_col = np.full(n_samples, np.nan)
    Rmse_col_down = np.full(n_samples, np.nan)
    Est_Lag_col = np.full(n_samples, np.nan)
    Est_Lag_col_down = np.full(n_samples, np.nan)
    end_time_col = np.full(n_samples, np.nan)
    end_time_col_down = np.full(n_samples, np.nan)
    time_step_col = np.full(n_samples, np.nan)
    time_step_col_down = np.full(n_samples, np.nan)
    elapsed_time_col = np.full(n_samples, np.nan)
    elapsed_time_col_down = np.full(n_samples, np.nan)
    Power_Ratio_ANN_BIC_col = np.full(n_samples, np.nan)
    Power_Ratio_ANN_BIC_col_down = np.full(n_samples, np.nan)
    Power_Ratio_ANN_TRI_col = np.full(n_samples, np.nan)
    Power_Ratio_ANN_TRI_col_down = np.full(n_samples, np.nan)
    T_col = np.full(n_samples, np.nan)
    T_col_down = np.full(n_samples, np.nan)

    Excitation = np.full(n_samples, np.nan)
    Activation = np.full(n_samples, np.nan)
    FiberLength = np.full(n_samples, np.nan)
    PennationAngle = np.full(n_samples, np.nan)
    CosPennationAngle = np.full(n_samples, np.nan)
    TendonLength = np.full(n_samples, np.nan)
    NormalizedFiberLength = np.full(n_samples, np.nan)
    FiberLengthAlongTendon = np.full(n_samples, np.nan)
    TendonStrain = np.full(n_samples, np.nan)
    passive_force_multiplier = np.full(n_samples, np.nan)
    ActiveForceLengthMultiplier = np.full(n_samples, np.nan)
    FiberVelocity = np.full(n_samples, np.nan)
    NormalizedFiberVelocity = np.full(n_samples, np.nan)
    FiberVelocityAlongTendon = np.full(n_samples, np.nan)
    TendonVelocity = np.full(n_samples, np.nan)
    ForceVelocityMultiplier = np.full(n_samples, np.nan)
    PennationAngularVelocity = np.full(n_samples, np.nan)
    FiberForce = np.full(n_samples, np.nan)
    FiberForceAlongTendon = np.full(n_samples, np.nan)
    ActiveFiberForce = np.full(n_samples, np.nan)
    PassiveFiberForce = np.full(n_samples, np.nan)
    ActiveFiberForceAlongTendon = np.full(n_samples, np.nan)
    PassiveFiberForceAlongTendon = np.full(n_samples, np.nan)
    TendonForce = np.full(n_samples, np.nan)
    FiberStiffness = np.full(n_samples, np.nan)
    FiberStiffnessAlongTendon = np.full(n_samples, np.nan)
    TendonStiffness = np.full(n_samples, np.nan)
    MuscleStiffness = np.full(n_samples, np.nan)
    FiberActivePower = np.full(n_samples, np.nan)
    FiberPassivePower = np.full(n_samples, np.nan)
    TendonPower = np.full(n_samples, np.nan)
    MusclePower = np.full(n_samples, np.nan)

    # 追従実験時の目標角度波形
    if TargetFlag:
        print("目標軌道作成中")
        time_array = np.arange(start_time, end_time + TIME_STEP * 6, TIME_STEP)
        target_angles, target_velocities, target_accelerations = Target_Angle(target_pattern, time_array)
    # ANNによる予測
    if control_condition == 6:  
        print("ANN(FF)による予測中")
        # モデルの読み込み
        ANNmodel = keras.models.load_model('{}/pre_learned_model.keras'.format(model_folder_name))
        inputs = []
        for i in range(len(target_angles) - 5):
            angle = np.degrees(target_angles[i:i+6])
            angular_velocity = np.degrees(target_velocities[i:i+6])
            angular_acceleration = np.degrees(target_accelerations[i:i+6])
            # 入力データの正規化
            # CustomScalerのインスタンス化
            custom_scaler_inputs_angle = CustomScaler(10, 90)
            custom_scaler_inputs_velocity = CustomScaler(-68.3, 68.3)
            custom_scaler_inputs_acceleration = CustomScaler(-180, 180)
            # 正規化
            angle = custom_scaler_inputs_angle.fit_transform(angle)
            angular_velocity = custom_scaler_inputs_velocity.fit_transform(angular_velocity)
            angular_acceleration = custom_scaler_inputs_acceleration.fit_transform(angular_acceleration)

            input_data = np.hstack([angle, angular_velocity, angular_acceleration])
            inputs.append(input_data)
        inputs = np.array(inputs)
        # 値の予測
        FF_output = ANNmodel.predict(inputs)
        FF_output_BIC = np.full(n_samples, np.nan)
        FF_output_TRI = np.full(n_samples, np.nan)
        for i in range (n_samples):
            FF_output_BIC[i] = FF_output[i][0]
            FF_output_TRI[i] = FF_output[i][1]
        output_scaler_BIC = CustomInverseScaler(-0.1, excitation_max_min_BIC)
        output_scaler_TRI = CustomInverseScaler(-0.1, excitation_max_min_TRI)
        # 予測結果を元のスケールに逆変換（出力データ用スケーラーを使用）
        FF_output_BIC = output_scaler_BIC.inverse_transform(FF_output_BIC)
        FF_output_TRI = output_scaler_TRI.inverse_transform(FF_output_TRI)
            
    # 制御器を仮作成
    # BIClong用
    controller_BIClong = osim.PrescribedController()
    controller_BIClong.setName("controller_BIClong")
    excitation_function_BIClong = osim.Constant(EXCITATION_MIN)
    controller_BIClong.addActuator(BIClong)
    controller_BIClong.prescribeControlForActuator(BIClong.getName(), excitation_function_BIClong)
    osimmodel.addController(controller_BIClong)
    # BICshort用
    controller_BICshort = osim.PrescribedController()
    controller_BICshort.setName("controller_BICshort")
    excitation_function_BICshort = osim.Constant(EXCITATION_MIN )
    controller_BICshort.addActuator(BICshort)
    controller_BICshort.prescribeControlForActuator(BICshort.getName(), excitation_function_BICshort)
    osimmodel.addController(controller_BICshort)
    # TRIlong用
    controller_TRIlong = osim.PrescribedController()
    controller_TRIlong.setName("controller_TRIlong")
    excitation_function_TRIlong = osim.Constant(EXCITATION_MIN )
    controller_TRIlong.addActuator(TRIlong)
    controller_TRIlong.prescribeControlForActuator(TRIlong.getName(), excitation_function_TRIlong)
    osimmodel.addController(controller_TRIlong)
    # TRIlat用
    controller_TRIlat = osim.PrescribedController()
    controller_TRIlat.setName("controller_TRIlat")
    excitation_function_TRIlat = osim.Constant(EXCITATION_MIN )
    controller_TRIlat.addActuator(TRIlat)
    controller_TRIlat.prescribeControlForActuator(TRIlat.getName(), excitation_function_TRIlat)
    osimmodel.addController(controller_TRIlat)
    # TRImed用
    controller_TRImed = osim.PrescribedController()
    controller_TRImed.setName("controller_TRImed")
    excitation_function_TRImed = osim.Constant(EXCITATION_MIN )
    controller_TRImed.addActuator(TRImed)
    controller_TRImed.prescribeControlForActuator(TRImed.getName(), excitation_function_TRImed)
    osimmodel.addController(controller_TRImed)
    # Bra用
    controller_Bra = osim.PrescribedController()
    controller_Bra.setName("controller_Bra")
    excitation_function_Bra = osim.Constant(EXCITATION_MIN )
    controller_Bra.addActuator(Bra)
    controller_Bra.prescribeControlForActuator(Bra.getName(), excitation_function_Bra)
    osimmodel.addController(controller_Bra)

    # CoordinateLimitForceを設定(肘関節の可動域設定)
    elbow_limit_force = osim.CoordinateLimitForce("r_elbow_flex", 150, 5, 1.5, 5, 0.1, 0.001)
    osimmodel.addForce(elbow_limit_force)

    # モデルに各種設定を反映
    state = osimmodel.initSystem()

    # シミュレーションオプション
    manager = osim.Manager(osimmodel)
    manager.setIntegratorMethod(osim.Manager.IntegratorMethod_RungeKuttaMerson)
    manager.setIntegratorMaximumStepSize(TIME_STEP)
    manager.setIntegratorMinimumStepSize(TIME_STEP)
    manager.initialize(state)
    state.setTime(start_time)

    # シミュレーション開始時の初期設定
    TIME_SIMULATION_START = time.time()
    if TargetFlag:
        LastError_BIC = target_angles[0] - INITIAL_ANGLE_RAD
        LastIntegralError_BIC = 0
        LastError_TRI = target_angles[0] - INITIAL_ANGLE_RAD
        LastIntegralError_TRI = 0

    # グラフ描画用設定
    if plot_flag:
        plt.ion()
        fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(10, 8))  # 2つのプロット領域を作成
        # シミュレーション中のデータ格納用リスト
        current_times = []
        state_values1 = []
        state_values2 = []
        state_values3 = []
        if TargetFlag:
            state_values4 = []

    # シミュレーション実行部分
    # 保存されるデータは時刻iにおいて積分後のデータ
    for i in range (n_samples):
        current_time = TIME_STEP * i
        # 初期時刻(t=0.0000)時の動作
        if current_time == 0:
            # 初期状態保存
            Time[i] = current_time
            Excitation_BIC[i] = EXCITATION_MIN
            Excitation_TRI[i] = EXCITATION_MIN
            Activation_BIC[i] = 0
            ObservedAngle_rad[i] = osimmodel.getStateVariableValues(state).get(2)
            ObservedAngularVelocity_rad[i] = osimmodel.getStateVariableValues(state).get(3)
            if TargetFlag:
                if control_condition == 6:
                    Excitation_PID_BIC[i] = 0
                    Excitation_ANN_BIC[i] = 0
                    TrainingData_Excitation_BIC[i] = 0
                    Excitation_PID_TRI[i] = 0
                    Excitation_ANN_TRI[i] = 0
                    TrainingData_Excitation_TRI[i] = 0
                Error_rad[i] = target_angles[0] - INITIAL_ANGLE_RAD
            if not iteration_flag:
                print("{}シミュレーション中:{}/{}".format(controller_name, round(current_time,2), end_time))
            elif iteration_flag:
                print("{}回/{}回 {}シミュレーション中:{}/{}".format(iteration, total_iteration, controller_name, round(current_time,2), end_time))
            # グラフ描画用
            if plot_flag:
                current_times.append(current_time)
                state_values1.append(np.degrees(ObservedAngle_rad[i]))
                state_values2.append(Excitation_BIC[i])
                state_values3.append(Excitation_TRI[i])
                if TargetFlag:
                    state_values4.append(np.degrees(target_angles[i]))
            continue
        # 制御器の設定
        if control_condition == 0: #Step
            if current_time < step_start:
                excitation_BIC = EXCITATION_MIN
                excitation_TRI = EXCITATION_MIN
            elif current_time <= end_time:
                excitation_BIC = excitation_BIC_max
                excitation_TRI = excitation_TRI_max
            else:
                excitation_BIC = EXCITATION_MIN
                excitation_TRI = EXCITATION_MIN
        if control_condition == 1: #Ramp
            if current_time < ramp_start:
                excitation_BIC = EXCITATION_MIN
                excitation_TRI = EXCITATION_MIN
            elif current_time <= ramp_end:
                excitation_BIC = EXCITATION_MIN  + (current_time - ramp_start) * (excitation_BIC_max - EXCITATION_MIN ) / (ramp_end - ramp_start)
                excitation_TRI = EXCITATION_MIN  + (current_time - ramp_start) * (excitation_TRI_max - EXCITATION_MIN ) / (ramp_end - ramp_start)
            else:
                excitation_BIC = EXCITATION_MIN
                excitation_TRI = EXCITATION_MIN
        if control_condition == 2: #PID
            target_angle = target_angles[i]
            if not excitation_frequency_flag:
                excitation_BIC, LastError_BIC, LastIntegralError_BIC = PID_BIC(target_angle, ObservedAngle_rad[i-1], LastError_BIC, LastIntegralError_BIC, excitation_ANN_BIC=0)
                excitation_TRI, LastError_TRI, LastIntegralError_TRI = PID_TRI(target_angle, ObservedAngle_rad[i-1], LastError_TRI, LastIntegralError_TRI, excitation_ANN_TRI=0)
            else:
            # 刺激周波数を考慮したバージョン
                if not square_wave_flag:
                    if i % int(1/excitation_frequency/TIME_STEP) == 1:
                        excitation_BIC, LastError_BIC, LastIntegralError_BIC = PID_BIC(target_angle, ObservedAngle_rad[i-1], LastError_BIC, LastIntegralError_BIC, excitation_ANN_BIC=0)
                        excitation_TRI, LastError_TRI, LastIntegralError_TRI = PID_TRI(target_angle, ObservedAngle_rad[i-1], LastError_TRI, LastIntegralError_TRI, excitation_ANN_TRI=0)
            # 刺激を矩形波にしたバージョン
                else:
                    if i % int(1/excitation_frequency/TIME_STEP) == 1:
                        excitation_BIC, LastError_BIC, LastIntegralError_BIC = PID_BIC(target_angle, ObservedAngle_rad[i-1], LastError_BIC, LastIntegralError_BIC, excitation_ANN_BIC=0)
                        excitation_TRI, LastError_TRI, LastIntegralError_TRI = PID_TRI(target_angle, ObservedAngle_rad[i-1], LastError_TRI, LastIntegralError_TRI, excitation_ANN_TRI=0)
                    if i % int(1/excitation_frequency/TIME_STEP) == (1+int(pulse_width/100)):
                        excitation_BIC = EXCITATION_MIN
                        excitation_TRI = EXCITATION_MIN
        if control_condition == 3: #Ramp2Step
            if current_time < ramp_start:
                excitation_BIC = EXCITATION_MIN 
            elif current_time <= ramp2step:
                excitation_BIC = EXCITATION_MIN  + (current_time - ramp_start) * (excitation_BIC_max - EXCITATION_MIN ) / (ramp2step - ramp_start)
            else:
                excitation_BIC = excitation_BIC_max
        if control_condition == 5: # Sin_Wave
            sec = end_time - sin_start
            t = np.arange(0, sec, TIME_STEP) #サンプリング点の生成
            y = np.sin(2*PI*frequency*t) # 正弦波の生成
            if current_time < ramp_start:
                excitation_BIC = EXCITATION_MIN 
            elif current_time <= ramp2step:
                excitation_BIC = EXCITATION_MIN  + (current_time - ramp_start) * (zero_point - EXCITATION_MIN ) / (ramp2step - ramp_start)
            elif current_time <= sin_start:
                excitation_BIC = zero_point
            else:
                excitation_BIC = zero_point + amplitude * y[i - int(7 / TIME_STEP) - 1]
        if control_condition == 6: # Feedback-Error-Learning controller
            excitation_ANN_BIC = FF_output_BIC[i]
            excitation_ANN_TRI = FF_output_TRI[i]
            excitation_PID_BIC, LastError_BIC, LastIntegralError_BIC = PID_BIC(target_angles[i], ObservedAngle_rad[i-1], LastError_BIC, LastIntegralError_BIC, excitation_ANN_BIC)
            excitation_PID_TRI, LastError_TRI, LastIntegralError_TRI = PID_TRI(target_angles[i], ObservedAngle_rad[i-1], LastError_TRI, LastIntegralError_TRI, excitation_ANN_TRI)
            trainingdata_excitation_BIC = excitation_PID_BIC + excitation_ANN_BIC
            trainingdata_excitation_TRI = excitation_PID_TRI + excitation_ANN_TRI
            excitation_BIC = excitation_PID_BIC + excitation_ANN_BIC + EXCITATION_MIN # Offset加算
            excitation_TRI = excitation_PID_TRI + excitation_ANN_TRI + EXCITATION_MIN # Offset加算
        if control_condition == 7: # 三角波
            if current_time < ramp_start:
                excitation_BIC = EXCITATION_MIN
                excitation_TRI = EXCITATION_MIN
            elif current_time <= switch_time+ramp_start:
                excitation_BIC = EXCITATION_MIN  + (current_time - ramp_start) * (excitation_BIC_max - EXCITATION_MIN ) / (switch_time+ramp_start - ramp_start)
                excitation_TRI = EXCITATION_MIN  + (current_time - ramp_start) * (excitation_TRI_max - EXCITATION_MIN ) / (switch_time+ramp_start - ramp_start)
            elif current_time <= switch_time*2+ramp_start:
                excitation_BIC = excitation_BIC_max + (-1) * (current_time - (switch_time+ramp_start)) * (excitation_BIC_max - EXCITATION_MIN ) / (switch_time*2+ramp_start - (switch_time+ramp_start))
                excitation_TRI = excitation_BIC_max + (-1) * (current_time - (switch_time+ramp_start)) * (excitation_TRI_max - EXCITATION_MIN ) / (switch_time*2+ramp_start - (switch_time+ramp_start))
            else:
                excitation_BIC = EXCITATION_MIN
                excitation_TRI = EXCITATION_MIN
        
        if square_wave_flag:
            excitation_BIC *= square_wave[i]

        # Limiter
        if excitation_BIC > excitation_BIC_max:
            excitation_BIC = excitation_BIC_max
        elif excitation_BIC < EXCITATION_MIN:
            excitation_BIC = EXCITATION_MIN
        if excitation_TRI > excitation_TRI_max:
            excitation_TRI = excitation_TRI_max
        elif excitation_TRI < EXCITATION_MIN:
            excitation_TRI = EXCITATION_MIN
        # 制御器に制御値をセット
        if target_muscle == 0 or target_muscle == 2:
            excitation_function_BIClong = osim.Constant(excitation_BIC)
            excitation_function_BICshort = osim.Constant(excitation_BIC)
            controller_BIClong.prescribeControlForActuator("BIClong", excitation_function_BIClong)
            controller_BICshort.prescribeControlForActuator("BICshort", excitation_function_BICshort)
            if target_muscle == 0:
                excitation_TRI = EXCITATION_MIN
        if target_muscle == 1 or target_muscle == 2:
            excitation_function_TRIlong = osim.Constant(excitation_TRI)
            excitation_function_TRIlat = osim.Constant(excitation_TRI)
            excitation_function_TRImed = osim.Constant(excitation_TRI)
            controller_TRIlong.prescribeControlForActuator("TRIlong", excitation_function_TRIlong)
            controller_TRIlat.prescribeControlForActuator("TRIlat", excitation_function_TRIlat)
            controller_TRImed.prescribeControlForActuator("TRImed", excitation_function_TRImed)
            if target_muscle == 1:
                excitation_BIC = EXCITATION_MIN

        # time_stepを一つ進めて積分実行
        state.setTime(current_time)
        state = manager.integrate(current_time + TIME_STEP)
        # データ保存
        Time[i] = current_time
        Excitation_BIC[i] = excitation_BIC
        Excitation_TRI[i] = excitation_TRI
        Activation_BIC[i] = osimmodel.getStateVariableValues(state).get(10)
        ObservedAngle_rad[i] = osimmodel.getStateVariableValues(state).get(2)
        ObservedAngularVelocity_rad[i] = osimmodel.getStateVariableValues(state).get(3)
        if TargetFlag:
            Error_rad[i] = target_angles[i] - osimmodel.getStateVariableValues(state).get(2)
        if control_condition == 6:
            Excitation_PID_BIC[i] = excitation_PID_BIC
            Excitation_ANN_BIC[i] = excitation_ANN_BIC
            TrainingData_Excitation_BIC[i] = trainingdata_excitation_BIC
            Excitation_PID_TRI[i] = excitation_PID_TRI
            Excitation_ANN_TRI[i] = excitation_ANN_TRI
            TrainingData_Excitation_TRI[i] = trainingdata_excitation_TRI
        if not iteration_flag:
            print("{}シミュレーション中:{}/{}".format(controller_name, round(current_time,2), end_time))
        elif iteration_flag:
            print("{}回/{}回 {}シミュレーション中:{}/{}".format(iteration, total_iteration, controller_name, round(current_time,2), end_time))
        
        Excitation[i] = BIClong.getExcitation(state)
        Activation[i] = BIClong.getActivation(state)
        FiberLength[i] = TRIlong.getFiberLength(state)
        PennationAngle[i] = BIClong.getPennationAngle(state)
        CosPennationAngle[i] = BIClong.getCosPennationAngle(state)
        TendonLength[i] = TRIlong.getTendonLength(state)
        NormalizedFiberLength[i] = BIClong.getNormalizedFiberLength(state)
        FiberLengthAlongTendon[i] = BIClong.getFiberLengthAlongTendon(state)
        TendonStrain[i] = BIClong.getTendonStrain(state)
        passive_force_multiplier[i] = BIClong.getPassiveForceMultiplier(state)
        ActiveForceLengthMultiplier[i] = BIClong.getActiveForceLengthMultiplier(state)
        FiberVelocity[i] = BIClong.getFiberVelocity(state)
        NormalizedFiberVelocity[i] = BIClong.getNormalizedFiberVelocity(state)
        FiberVelocityAlongTendon[i] = BIClong.getFiberVelocityAlongTendon(state)
        TendonVelocity[i] = BIClong.getTendonVelocity(state)
        ForceVelocityMultiplier[i] = BIClong.getForceVelocityMultiplier(state)
        PennationAngularVelocity[i] = BIClong.getPennationAngularVelocity(state)
        FiberForce[i] = BIClong.getFiberForce(state)
        FiberForceAlongTendon[i] = BIClong.getFiberForceAlongTendon(state)
        ActiveFiberForce[i] = BIClong.getActiveFiberForce(state)
        PassiveFiberForce[i] = BIClong.getPassiveFiberForce(state)
        ActiveFiberForceAlongTendon[i] = BIClong.getActiveFiberForceAlongTendon(state)
        PassiveFiberForceAlongTendon[i] = BIClong.getPassiveFiberForceAlongTendon(state)
        TendonForce[i] = BIClong.getTendonForce(state)
        FiberStiffness[i] = BIClong.getFiberStiffness(state)
        FiberStiffnessAlongTendon[i] = BIClong.getFiberStiffnessAlongTendon(state)
        TendonStiffness[i] = BIClong.getTendonStiffness(state)
        MuscleStiffness[i] = BIClong.getMuscleStiffness(state)
        FiberActivePower[i] = BIClong.getFiberActivePower(state)
        FiberPassivePower[i] = BIClong.getFiberPassivePower(state)
        TendonPower[i] = BIClong.getTendonPower(state)
        MusclePower[i] = BIClong.getMusclePower(state)

        # グラフ描画用
        if plot_flag:
            current_times.append(current_time)
            state_values1.append(np.degrees(ObservedAngle_rad[i]))
            state_values2.append(Excitation_BIC[i])
            state_values3.append(Excitation_TRI[i])
            if TargetFlag:
                state_values4.append(np.degrees(target_angles[i]))
            if i % 500 == 0:
                ax1.cla()  # グラフクリア
                ax1.plot(current_times, state_values1, label='Observed Angle')  # 新しいデータでプロット
                if TargetFlag:
                    ax1.plot(current_times, state_values4, label='Target Angle')  # 新しいデータでプロット
                ax1.set_xlabel('Time [sec]')
                ax1.set_ylabel('Elbow Joint Angle [deg]')
                ax1.legend()
                ax2.cla()  # グラフクリア
                ax2.plot(current_times, state_values2, label='Excitation_BIC')  # 新しいデータでプロット
                ax2.plot(current_times, state_values3, label='Excitation_TRI')  # 新しいデータでプロット
                ax2.set_xlabel('Time [sec]')
                ax2.set_ylabel('Excitation Intensity')
                ax2.legend()
                plt.draw()
                plt.pause(0.01)
    # シミュレーション終了時の動作
    TIME_SIMULATION_END = time.time()
    elapsed_time = TIME_SIMULATION_END - TIME_SIMULATION_START
    print("シミュレーション終了")
    print(f"シミュレーション時間: {elapsed_time} 秒")

    # データ格納
    ObservedAngle_degree = np.degrees(ObservedAngle_rad)
    ObservedAngularVelocity_degree = np.degrees(ObservedAngularVelocity_rad)
    ObservedAngularAcceleration_degree[0] = 0
    for i in range(n_samples): # 角加速度
        if i == 0:
            ObservedAngularAcceleration_degree[i] = 0
        else:
            ObservedAngularAcceleration_degree[i] = (ObservedAngularVelocity_degree[i] - ObservedAngularVelocity_degree[i-1]) / TIME_STEP
    if TargetFlag:
        for i in range(n_samples):
            TargetAngle_degree[i] = np.degrees(target_angles[i])
            TargetAngularVelocity_degree[i] = np.degrees(target_velocities[i])
            TargetAngularAcceleration_degree[i] = np.degrees(target_accelerations[i])
        Error_degree = np.degrees(Error_rad)
        Mae = np.average(np.abs(Error_degree[for_sec_index:]))
        Mae_col[0] = Mae
        Rmse = np.sqrt(np.average(np.square(Error_degree[for_sec_index:])))
        print(Rmse)
        Rmse_col[0] = Rmse
        Est_Lag_col[0] = calculate_estimation_lag(TargetAngle_degree, ObservedAngle_degree, TIME_STEP)
    end_time_col[0] = end_time
    time_step_col[0] = TIME_STEP
    elapsed_time_col[0] = elapsed_time
    T_col[0] = T1
    T_col[1] = T2
    T_col[2] = T3

    if max(ObservedAngle_degree) > 180 or min(ObservedAngle_degree) < 0:
        bug_flag = True

    if control_condition == 6:
        for i in range(len(Excitation_PID_BIC) - for_sec_index):
            Power_Ratio_ANN_BIC += 100 * (Excitation_ANN_BIC[i + for_sec_index] ** 2) / (Excitation_ANN_BIC[i + for_sec_index] ** 2 + Excitation_PID_BIC[i + for_sec_index] ** 2)
            Power_Ratio_ANN_TRI += 100 * (Excitation_ANN_TRI[i + for_sec_index] ** 2) / (Excitation_ANN_TRI[i + for_sec_index] ** 2 + Excitation_PID_TRI[i + for_sec_index] ** 2)
        Power_Ratio_ANN_BIC /= (len(Excitation_PID_BIC) - for_sec_index)
        Power_Ratio_ANN_TRI /= (len(Excitation_PID_BIC) - for_sec_index)
        Power_Ratio_ANN_BIC_col[0] = Power_Ratio_ANN_BIC
        Power_Ratio_ANN_TRI_col[0] = Power_Ratio_ANN_TRI
        if iteration >= 11 and (iteration - 1) % 10 == 0:
            if (np.average(training_history[5:10]) - np.average(training_history[0:5])) < 0.5:
                training_history[10] += 1
            if (np.average(training_history[16:21]) - np.average(training_history[11:16])) < 0.5:
                training_history[21] += 1
        training_history[(iteration - 1) % 10] = Power_Ratio_ANN_BIC
        training_history[(iteration - 1) % 10 + 11] = Power_Ratio_ANN_TRI
        
    #if not control_condition == 6 or (control_condition == 6 and iteration % 10 == 1):
    if not control_condition == 6 or control_condition == 6:
        # 以下、excelへシミュレーション結果の保存
        print("結果保存中")
        # ダウンサンプリング
        for i in range(n_samples_down):
            Time_down[i] = Time[i*downsamplerate]
            Excitation_BIC_down[i] = Excitation_BIC[i*downsamplerate]
            Excitation_PID_BIC_down[i] = Excitation_PID_BIC[i*downsamplerate]
            Excitation_ANN_BIC_down[i] = Excitation_ANN_BIC[i*downsamplerate]
            TrainingData_Excitation_BIC_down[i] = TrainingData_Excitation_BIC[i*downsamplerate]
            Excitation_TRI_down[i] = Excitation_TRI[i*downsamplerate]
            Excitation_PID_TRI_down[i] = Excitation_PID_TRI[i*downsamplerate]
            Excitation_ANN_TRI_down[i] = Excitation_ANN_TRI[i*downsamplerate]
            TrainingData_Excitation_TRI_down[i] = TrainingData_Excitation_TRI[i*downsamplerate]
            Activation_down[i] = Activation_BIC[i*downsamplerate]
            ObservedAngle_rad_down[i] = ObservedAngle_rad[i*downsamplerate]
            ObservedAngle_degree_down[i] = ObservedAngle_degree[i*downsamplerate]
            ObservedAngularVelocity_degree_down[i] = ObservedAngularVelocity_degree[i*downsamplerate]
            ObservedAngularAcceleration_degree_down[i] = ObservedAngularVelocity_degree[i*downsamplerate]
            TargetAngle_degree_down[i] = TargetAngle_degree[i*downsamplerate]
            Error_rad_down[i] = Error_rad[i*downsamplerate]
            Error_degree_down[i] = Error_degree[i*downsamplerate]
            Mae_col_down[i] = Mae_col[i*downsamplerate]
            Rmse_col_down[i] = Rmse_col[i*downsamplerate]
            Est_Lag_col_down[i] = Est_Lag_col[i*downsamplerate]
            end_time_col_down[i] = end_time_col[i*downsamplerate]
            time_step_col_down[i] = time_step_col[i*downsamplerate]
            elapsed_time_col_down[i] = elapsed_time_col[i*downsamplerate]
            Power_Ratio_ANN_BIC_col_down[i] = Power_Ratio_ANN_BIC_col[i*downsamplerate]
            Power_Ratio_ANN_TRI_col_down[i] = Power_Ratio_ANN_TRI_col[i*downsamplerate]
        # ダウンサンプリング前のデータフレーム
        data = {
            "Time": Time,
            "Excitation_BIC": Excitation_BIC,
            "Excitation_PID_BIC": Excitation_PID_BIC,
            "Excitation_ANN_BIC": Excitation_ANN_BIC,
            "TrainingData_Excitation_BIC": TrainingData_Excitation_BIC,
            "Excitation_TRI": Excitation_TRI,
            "Excitation_PID_TRI": Excitation_PID_TRI,
            "Excitation_ANN_TRI": Excitation_ANN_TRI,
            "TrainingData_Excitation_TRI": TrainingData_Excitation_TRI,
            "ObservedAngle": ObservedAngle_degree,
            "TargetAngle": TargetAngle_degree,
            "ObservedAngularVelocity": ObservedAngularVelocity_degree,
            "TargetAngularVelocity": TargetAngularVelocity_degree,
            "ObservedAngularAcceleration": ObservedAngularAcceleration_degree,
            "TargetAngularAcceleration": TargetAngularAcceleration_degree,
            "end_time": end_time_col,
            "TIME_STEP": time_step_col,
            "elapsed_time": elapsed_time_col,
            "Power_Ratio_ANN_BIC": Power_Ratio_ANN_BIC_col,
            "Power_Ratio_ANN_TRI": Power_Ratio_ANN_TRI_col,
            "MAE": Mae_col,
            "RMSE": Rmse_col,
            "Est_Lag": Est_Lag_col,
            "T": T_col,
            "Excitation":Excitation,
            "Activation":Activation,
            "FiberLength":FiberLength,
            "PennationAngle":PennationAngle,
            "CosPennationAngle":CosPennationAngle,
            "TendonLength":TendonLength,
            "NormalizedFiberLength":NormalizedFiberLength,
            "FiberLengthAlongTendon":FiberLengthAlongTendon,
            "TendonStrain":TendonStrain,
            "passive_force_multiplier":passive_force_multiplier,
            "ActiveForceLengthMultiplier":ActiveForceLengthMultiplier,
            "FiberVelocity":FiberVelocity,
            "NormalizedFiberVelocity":NormalizedFiberVelocity,
            "FiberVelocityAlongTendon":FiberVelocityAlongTendon,
            "TendonVelocity":TendonVelocity,
            "ForceVelocityMultiplier":ForceVelocityMultiplier,
            "PennationAngularVelocity":PennationAngularVelocity,
            "FiberForce":FiberForce,
            "FiberForceAlongTendon":FiberForceAlongTendon,
            "ActiveFiberForce":ActiveFiberForce,
            "PassiveFiberForce":PassiveFiberForce,
            "ActiveFiberForceAlongTendon":ActiveFiberForceAlongTendon,
            "PassiveFiberForceAlongTendon":PassiveFiberForceAlongTendon,
            "TendonForce":TendonForce,
            "FiberStiffness":FiberStiffness,
            "FiberStiffnessAlongTendon":FiberStiffnessAlongTendon,
            "TendonStiffness":TendonStiffness,
            "MuscleStiffness":MuscleStiffness,
            "FiberActivePower":FiberActivePower,
            "FiberPassivePower":FiberPassivePower,
            "TendonPower":TendonPower,
            "MusclePower":MusclePower,
        }
        df = pd.DataFrame(data)
        # ダウンサンプリング後のデータフレーム
        data_down = {
            "Time": Time_down,
            "Excitation_BIC": Excitation_BIC_down,
            "Excitation_PID_BIC": Excitation_PID_BIC_down,
            "Excitation_ANN_BIC": Excitation_ANN_BIC_down,
            "TrainingData_Excitation_BIC": TrainingData_Excitation_BIC_down,
            "Excitation_TRI": Excitation_TRI_down,
            "Excitation_PID_TRI": Excitation_PID_TRI_down,
            "Excitation_ANN_TRI": Excitation_ANN_TRI_down,
            "TrainingData_Excitation_TRI": TrainingData_Excitation_TRI_down,
            "ObservedAngle": ObservedAngle_degree_down,
            "TargetAngle": TargetAngle_degree_down,
            "ObservedAngularVelocity": ObservedAngularVelocity_degree_down,
            "TargetAngularVelocity": TargetAngularVelocity_degree_down,
            "ObservedAngularAcceleration": ObservedAngularAcceleration_degree_down,
            "TargetAngularAcceleration": TargetAngularAcceleration_degree_down,
            "end_time": end_time_col_down,
            "TIME_STEP": time_step_col_down,
            "elapsed_time": elapsed_time_col_down,
            "Power_Ratio_ANN_BIC": Power_Ratio_ANN_BIC_col_down,
            "Power_Ratio_ANN_TRI": Power_Ratio_ANN_TRI_col_down,
            "MAE": Mae_col_down,
            "RMSE": Rmse_col_down,
            "Est_Lag": Est_Lag_col_down,
            "T": T_col
        }
        df_down = pd.DataFrame(data_down)
        # 現在のディレクトリと新しいフォルダ名を結合して新しいパスを作成
        folder_path = os.path.join(os.getcwd(), result_folder_name)
        # フォルダが存在しない場合は作成
        if not os.path.exists(folder_path):
            os.makedirs(folder_path)
        # Excelファイルの保存パス
        excel_filename = os.path.join(folder_path, '{}_{}.xlsx'.format(controller_name, datetime.datetime.now().strftime("%m-%d_%H-%M-%S")))
        if iteration_flag:
            excel_filename = os.path.join(folder_path, '{}_{}_{}.xlsx'.format(iteration, controller_name, datetime.datetime.now().strftime("%m-%d_%H-%M-%S")))
        if bug_flag:
            excel_filename = os.path.join(folder_path, 'バグ{}_{}_{}.xlsx'.format(iteration, controller_name, datetime.datetime.now().strftime("%m-%d_%H-%M-%S")))
        # Excelにデータ保存
        with pd.ExcelWriter(excel_filename) as writer:
            df.to_excel(writer, sheet_name='before_down', index=False)
            df_down.to_excel(writer, sheet_name='after_down', index=False)
        print("保存完了")

    if control_condition == 6 and not bug_flag:
        # Excelファイルへのパスを定義
        result_dir = os.path.join(os.getcwd(), result_folder_name)
        if not os.path.exists(result_dir):
            os.makedirs(result_dir)
        file_path = os.path.join(result_dir, 'history.xlsx')
        
        # Excelファイルに保存するデータを準備
        data = {
            'Iteration': [iteration],
            'Power_Ratio_ANN_BIC': [Power_Ratio_ANN_BIC],
            'Power_Ratio_ANN_TRI': [Power_Ratio_ANN_TRI],
            'MAE': [Mae],
            'RMSE': [Rmse],
            'lr_BIC': [training_history[10]],
            'lr_TRi': [training_history[21]]
        }
        df_new = pd.DataFrame(data)
        
        # ファイルが既に存在するか確認し、適切に処理
        if os.path.exists(file_path):
            # ファイルが存在する場合、読み込んで新しいデータを追加
            df_existing = pd.read_excel(file_path)
            df_updated = pd.concat([df_existing, df_new], ignore_index=True)
        else:
            # ファイルが存在しない場合、新しいファイルを作成
            df_updated = df_new
        
        # Excelファイルにデータを保存
        with pd.ExcelWriter(file_path, engine='openpyxl', mode='w') as writer:
            df_updated.to_excel(writer, index=False)

    # pre_learned_model.kerasの再学習
    if ANN_training_flag:
        if control_condition == 6 and not bug_flag < 0:
            # 出力データの正規化
            custom_scaler_outputs_BIC = CustomScaler(-0.1, excitation_max_min_BIC)
            custom_scaler_outputs_TRI = CustomScaler(-0.1, excitation_max_min_TRI)
            outputs_BIC = custom_scaler_outputs_BIC.fit_transform(TrainingData_Excitation_BIC)
            outputs_TRI = custom_scaler_outputs_TRI.fit_transform(TrainingData_Excitation_TRI)
            outputs = np.vstack([outputs_BIC[for_sec_index:], outputs_TRI[for_sec_index:]]).T
            inputs = inputs[for_sec_index:]
            # モデルのコンパイル
            lr = 0.001 / (2 ** training_history[10])
            ANNmodel.compile(optimizer=keras.optimizers.Adam(learning_rate=0.001), loss='mean_squared_error')
            # モデルの学習
            ANNmodel.fit(inputs, outputs, epochs=1, batch_size=250)
            # モデルの保存
            model_save_path = "{}/pre_learned_model.keras".format(model_folder_name)
            ANNmodel.save(model_save_path)
            print(f"Model saved to {model_save_path}")

    if plot_flag:
        plt.ioff()
        plt.show(block=False)  # グラフを非ブロッキングモードで表示
        time.sleep(0.5)  # 5秒間待機
        plt.close()  # グラフウィンドウを閉じる

    return bug_flag, training_history