"""
OpenSimでのシミュレーション結果を、ランプ応答について、運動閾値と最大角度間で抜き出し、また正規化する
フォルダーパスとファイル名を指定して実行
運動閾値：初期角度+0.5°
最大角度：最終角度-0.5°

"""

import os
import pandas as pd
from sklearn.preprocessing import MinMaxScaler

# フォルダー名とCSVファイル名を指定
folder_path = r'\\Fesback\fes\Ryusuke_Yokoya\M2\osim\Results\0730\nonlinear'
csv_file_name = '√.csv'

# CSVファイルのフルパスを取得
csv_file_path = os.path.join(folder_path, csv_file_name)

# CSVファイルを読み込み
df = pd.read_csv(csv_file_path)

# 初期値と最終値を取得
initial_value = df['ObservedAngle'].iloc[0]
final_value = df['ObservedAngle'].iloc[-1]

# 閾値インデックスと終値インデックスを計算
threshold_index = df[df['ObservedAngle'] >= initial_value + 1].index[0]
end_index = df[df['ObservedAngle'] >= final_value - 1].index[0]

# 閾値インデックスから終値インデックスまでの値を抽出
time_extracted = df.loc[threshold_index:end_index, 'Time'].reset_index(drop=True)
excitation_bic_extracted = df.loc[threshold_index:end_index, 'Excitation_BIC'].reset_index(drop=True)
observed_angle_extracted = df.loc[threshold_index:end_index, 'ObservedAngle'].reset_index(drop=True)

# 抽出した値を新しい列として追加
df[f'Time_extracted'] = pd.Series(time_extracted)
df[f'Excitation_BIC_extracted'] = pd.Series(excitation_bic_extracted)
df[f'ObservedAngle_extracted'] = pd.Series(observed_angle_extracted)

# 正規化
scaler = MinMaxScaler()
excitation_bic_normalized = scaler.fit_transform(excitation_bic_extracted.values.reshape(-1, 1)).flatten()
observed_angle_normalized = scaler.fit_transform(observed_angle_extracted.values.reshape(-1, 1)).flatten()

# 正規化された値を新しい列として追加
df[f'Excitation_BIC_extracted_normalized'] = pd.Series(excitation_bic_normalized)
df[f'ObservedAngle_extracted_normalized'] = pd.Series(observed_angle_normalized)

# CSVファイルを上書き保存
df.to_csv(csv_file_path, index=False)

print(f"CSVファイル '{csv_file_name}' に新しい列が追加されました。")
