import tkinter as tk
from tkinter import filedialog
from tkinter import ttk
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

# デフォルトフォントを設定
plt.rcParams['font.family'] = 'DejaVu Sans'
initial_dir = r"\\Fesback\fes\Ryusuke_Yokoya\M2\osim\Results\リクルートメント実験結果"

class GraphPlotterApp:
    def __init__(self, root):
        self.root = root
        self.root.title("Graph Plotter")
        self.root.geometry("1000x600")  # 初期ウィンドウサイズを設定
        
        self.filepaths = []
        self.dataframes = {}
        
        self.create_widgets()
        
    def create_widgets(self):
        # タブコントロールの作成
        tab_control = ttk.Notebook(self.root)
        
        # 横軸共通タブの作成
        tab1 = ttk.Frame(tab_control)
        tab_control.add(tab1, text="横軸別々")
        
        tab_control.pack(expand=1, fill='both')

        # スクロール可能なキャンバスの作成
        canvas = tk.Canvas(tab1)
        scrollbar = ttk.Scrollbar(tab1, orient="vertical", command=canvas.yview)
        scrollable_frame = ttk.Frame(canvas)

        scrollable_frame.bind(
            "<Configure>",
            lambda e: canvas.configure(
                scrollregion=canvas.bbox("all")
            )
        )

        canvas.create_window((0, 0), window=scrollable_frame, anchor="nw")
        canvas.configure(yscrollcommand=scrollbar.set)

        canvas.pack(side="left", fill="both", expand=True)
        scrollbar.pack(side="right", fill="y")

        # 横軸共通タブ内のウィジェットを配置
        self.create_common_widgets(scrollable_frame)
        
    def create_common_widgets(self, parent):
        # メインフレーム
        main_frame = ttk.Frame(parent, padding="10 10 10 10")
        main_frame.grid(column=0, row=0, sticky=(tk.W, tk.E, tk.N, tk.S))
        
        # グリッドの設定
        main_frame.columnconfigure(0, weight=1)
        main_frame.columnconfigure(1, weight=1)
        main_frame.columnconfigure(2, weight=1)
        main_frame.columnconfigure(3, weight=1)
        main_frame.columnconfigure(4, weight=1)

        # ファイル選択ボタン
        self.file_select_button = ttk.Button(main_frame, text="ファイル選択", command=self.load_files)
        self.file_select_button.grid(row=0, column=0, padx=5, pady=5, sticky=tk.W)
        
        # プロットボタン
        self.plot_button = ttk.Button(main_frame, text="プロット", command=self.plot_graph)
        self.plot_button.grid(row=0, column=1, padx=5, pady=5, sticky=tk.W)
        
        # リセットボタン
        self.reset_button = ttk.Button(main_frame, text="リセット", command=self.reset_app)
        self.reset_button.grid(row=0, column=2, padx=5, pady=5, sticky=tk.W)
        
        # 選択されたファイルの一覧表示用フレーム
        self.file_list_frame = ttk.Frame(main_frame)
        self.file_list_frame.grid(row=1, column=0, columnspan=4, padx=5, pady=5, sticky=tk.W)
        
        # グラフタイトル
        ttk.Label(main_frame, text="グラフタイトル").grid(row=2, column=0, padx=5, pady=5, sticky=tk.E)
        self.title_entry = ttk.Entry(main_frame, width=30)
        self.title_entry.grid(row=2, column=1, padx=5, pady=5, sticky=tk.W)
        
        # 横軸ラベル
        ttk.Label(main_frame, text="横軸ラベル").grid(row=3, column=0, padx=5, pady=5, sticky=tk.E)
        self.x_label_entry = ttk.Entry(main_frame, width=30)
        self.x_label_entry.grid(row=3, column=1, padx=5, pady=5, sticky=tk.W)
        self.x_label_entry.insert(0, "Time [sec]")
        
        # 第二行目の横軸ラベル
        ttk.Label(main_frame, text="第二行目の横軸ラベル").grid(row=4, column=0, padx=5, pady=5, sticky=tk.E)
        self.x2_label_entry = ttk.Entry(main_frame, width=30)
        self.x2_label_entry.grid(row=4, column=1, padx=5, pady=5, sticky=tk.W)
        
        # 第一縦軸ラベル
        ttk.Label(main_frame, text="第一縦軸ラベル").grid(row=5, column=0, padx=5, pady=5, sticky=tk.E)
        self.y_label_entry = ttk.Entry(main_frame, width=30)
        self.y_label_entry.grid(row=5, column=1, padx=5, pady=5, sticky=tk.W)

        # 第二縦軸ラベル
        ttk.Label(main_frame, text="第二縦軸ラベル").grid(row=6, column=0, padx=5, pady=5, sticky=tk.E)
        self.y2_label_entry = ttk.Entry(main_frame, width=30)
        self.y2_label_entry.grid(row=6, column=1, padx=5, pady=5, sticky=tk.W)

        # グラフサイズ
        ttk.Label(main_frame, text="グラフサイズ (幅, 高さ)").grid(row=7, column=0, padx=5, pady=5, sticky=tk.E)
        self.size_entry = ttk.Entry(main_frame, width=30)
        self.size_entry.grid(row=7, column=1, padx=5, pady=5, sticky=tk.W)
        self.size_entry.insert(0, "(10, 6)")
        
        # 縦軸データ追加ボタン
        self.add_y_axis_button = ttk.Button(main_frame, text="データ追加", command=self.add_y_axis_section)
        self.add_y_axis_button.grid(row=7, column=2, padx=5, pady=5, sticky=tk.W)
        
        # 縦軸データセクションを追加するためのフレーム
        self.y_axis_frame = ttk.Frame(main_frame, padding="10 0 0 0")
        self.y_axis_frame.grid(row=8, column=0, columnspan=4, sticky=(tk.W, tk.E))
        
        self.y_axis_sections = []
        self.add_y_axis_section()
        
    def load_files(self):
        filepaths = filedialog.askopenfilenames(initialdir=initial_dir, filetypes=[("CSV files", "*.csv")])
        if filepaths:
            for filepath in filepaths:
                self.filepaths.append(filepath)
                df = pd.read_csv(filepath)
                self.dataframes[filepath] = df
            self.update_file_comboboxes()
            self.update_file_list()
        
    def update_file_comboboxes(self):
        files = [self.extract_filename(fp) for fp in self.dataframes.keys()]
        for section in self.y_axis_sections:
            section['file_combobox']['values'] = files
            
    def update_file_list(self):
        for widget in self.file_list_frame.winfo_children():
            widget.destroy()
        
        for filepath in self.filepaths:
            frame = ttk.Frame(self.file_list_frame)
            frame.pack(fill=tk.X, pady=2)
            
            remove_button = ttk.Button(frame, text="✕", width=2, command=lambda fp=filepath: self.remove_file(fp))
            remove_button.pack(side=tk.LEFT, padx=2)
            
            label = ttk.Label(frame, text=filepath)
            label.pack(side=tk.LEFT, padx=5)
            
    def remove_file(self, filepath):
        self.filepaths.remove(filepath)
        del self.dataframes[filepath]
        self.update_file_comboboxes()
        self.update_file_list()
            
    def extract_filename(self, filepath):
        return filepath.split('/')[-1]
            
    def update_columns(self, event, section):
        filename = section['file_combobox'].get()
        filepath = next(fp for fp in self.dataframes if self.extract_filename(fp) == filename)
        columns = list(self.dataframes[filepath].columns)
        section['x_column_combobox']['values'] = columns
        section['y_column_combobox']['values'] = columns
        
        if self.y_axis_sections[0] == section:
            for sec in self.y_axis_sections:
                sec['x_column_combobox'].set(section['x_column_combobox'].get())
        
    def add_y_axis_section(self):
        row_index = len(self.y_axis_sections)
        
        remove_button = ttk.Button(self.y_axis_frame, text="✕", width=2, command=lambda: self.remove_y_axis_section(section))
        remove_button.grid(row=row_index, column=0, padx=2, pady=5)
        
        use_secondary_y = tk.BooleanVar()
        secondary_checkbox = tk.Checkbutton(self.y_axis_frame, variable=use_secondary_y, width=2)
        secondary_checkbox.grid(row=row_index, column=1, padx=2, pady=5)
        
        plot_in_second_row = tk.BooleanVar()
        second_row_checkbox = tk.Checkbutton(self.y_axis_frame, variable=plot_in_second_row, width=2)
        second_row_checkbox.grid(row=row_index, column=2, padx=2, pady=5)
        
        file_combobox = ttk.Combobox(self.y_axis_frame, state="readonly", width=25)
        file_combobox.grid(row=row_index, column=3, padx=5, pady=5)
        
        x_column_combobox = ttk.Combobox(self.y_axis_frame, state="readonly", width=25)
        x_column_combobox.grid(row=row_index, column=4, padx=5, pady=5)
        
        y_column_combobox = ttk.Combobox(self.y_axis_frame, state="readonly", width=25)
        y_column_combobox.grid(row=row_index, column=5, padx=5, pady=5)
        
        color_combobox = ttk.Combobox(self.y_axis_frame, state="readonly", values=["赤", "青", "緑", "黄色", "黒", "オレンジ", "紫", "茶色", "ピンク", "灰色"], width=8)
        color_combobox.grid(row=row_index, column=6, padx=5, pady=5)
        
        linestyle_combobox = ttk.Combobox(self.y_axis_frame, state="readonly", values=["実線", "破線", "点線"], width=8)
        linestyle_combobox.set("実線")  # デフォルトを「実線」に設定
        linestyle_combobox.grid(row=row_index, column=7, padx=5, pady=5)

        legend_entry = ttk.Entry(self.y_axis_frame, width=15)
        legend_entry.grid(row=row_index, column=8, padx=5, pady=5)
        
        section = {
            'remove_button': remove_button,
            'secondary_checkbox': secondary_checkbox,
            'plot_in_second_row': plot_in_second_row,
            'use_secondary_y': use_secondary_y,
            'file_combobox': file_combobox,
            'x_column_combobox': x_column_combobox,
            'y_column_combobox': y_column_combobox,
            'color_combobox': color_combobox,
            'linestyle_combobox': linestyle_combobox,
            'legend_entry': legend_entry
        }
        
        file_combobox.bind("<<ComboboxSelected>>", lambda event, sec=section: self.update_columns(event, sec))
        
        self.y_axis_sections.append(section)
        self.update_file_comboboxes()
        
    def remove_y_axis_section(self, section):
        section['remove_button'].grid_forget()
        section['secondary_checkbox'].grid_forget()
        section['plot_in_second_row'].grid_forget()
        section['file_combobox'].grid_forget()
        section['x_column_combobox'].grid_forget()
        section['y_column_combobox'].grid_forget()
        section['color_combobox'].grid_forget()
        section['linestyle_combobox'].grid_forget()
        section['legend_entry'].grid_forget()
        self.y_axis_sections.remove(section)
        self.update_file_comboboxes()

    def reset_app(self):
        self.filepaths = []
        self.dataframes = {}
        self.y_axis_sections = []
        for widget in self.file_list_frame.winfo_children():
            widget.destroy()
        self.title_entry.delete(0, tk.END)
        self.x_label_entry.delete(0, tk.END)
        self.x_label_entry.insert(0, "Time [sec]")
        self.x2_label_entry.delete(0, tk.END)
        self.y_label_entry.delete(0, tk.END)
        self.y2_label_entry.delete(0, tk.END)
        self.size_entry.delete(0, tk.END)
        self.size_entry.insert(0, "(10, 6)")
        for widget in self.y_axis_frame.winfo_children():
            widget.destroy()
        self.add_y_axis_section()
        
    def plot_graph(self):
        size_text = self.size_entry.get()
        width, height = eval(size_text)  # Safe because we control the input format
        
        # チェックボックスが選択されているデータがあるかを確認
        second_row_needed = any(section['plot_in_second_row'].get() for section in self.y_axis_sections)
        
        if second_row_needed:
            fig, axs = plt.subplots(nrows=2, ncols=1, figsize=(width, height), sharex=False)
            second_row_ax = axs[1]
        else:
            fig, axs = plt.subplots(nrows=1, ncols=1, figsize=(width, height), sharex=False)
            axs = [axs]
        
        fig.tight_layout(pad=3.0)
        
        primary_ax = axs[0]
        secondary_ax_used = False
        second_row_ax_used = False
        
        color_dict = {"赤": "red", "青": "blue", "緑": "green", "黄色": "yellow", "黒": "black", "オレンジ": "orange", "紫": "purple", "茶色": "brown", "ピンク": "pink", "灰色": "gray"}
        linestyle_dict = {"実線": "-", "破線": "--", "点線": ":"}
        
        for section in self.y_axis_sections:
            file = section['file_combobox'].get()
            x_column = section['x_column_combobox'].get()
            y_column = section['y_column_combobox'].get()
            color = color_dict[section['color_combobox'].get()]
            linestyle = linestyle_dict[section['linestyle_combobox'].get()]
            legend = section['legend_entry'].get()
            use_secondary_y = section['use_secondary_y'].get()
            plot_in_second_row = section['plot_in_second_row'].get()
            
            filepath = next(fp for fp in self.dataframes if self.extract_filename(fp) == file)
            x_data = self.dataframes[filepath][x_column]
            y_data = self.dataframes[filepath][y_column]
            
            if second_row_needed and plot_in_second_row:
                second_row_ax_used = True
                second_row_ax.plot(x_data, y_data, color=color, linestyle=linestyle, label=legend)
            else:
                if use_secondary_y:
                    if not secondary_ax_used:
                        ax2 = primary_ax.twinx()
                        secondary_ax_used = True
                    ax2.plot(x_data, y_data, color=color, linestyle=linestyle, label=legend)
                else:
                    primary_ax.plot(x_data, y_data, color=color, linestyle=linestyle, label=legend)
        
        title = self.title_entry.get()
        x_label = self.x_label_entry.get()
        x2_label = self.x2_label_entry.get()
        y_label = self.y_label_entry.get()
        y2_label = self.y2_label_entry.get()
        
        if title:
            primary_ax.set_title(title)
        if x_label:
            primary_ax.set_xlabel(x_label)
        if y_label:
            primary_ax.set_ylabel(y_label)
        if secondary_ax_used and y2_label:
            ax2.set_ylabel(y2_label)
        if x2_label and second_row_ax_used:
            second_row_ax.set_xlabel(x2_label)

        primary_ax.legend(loc='upper left')
        if secondary_ax_used:
            ax2.legend(loc='upper right')
        if second_row_ax_used:
            second_row_ax.legend(loc='upper left')

        for ax in axs:
            if ax == primary_ax:
                if "Time" in x_column:
                    ax.set_xticks(range(int(min(x_data)), int(max(x_data))+1))
                elif "Excitation" in x_column and "Excitation_BIC_extracted_normalized" not in x_column:
                    ax.set_xticks(np.arange(min(x_data), max(x_data)+0.01, 0.01))
                    ax.xaxis.set_major_formatter(plt.FuncFormatter(lambda x, _: '{:.2f}'.format(x)))
                elif "Excitation_BIC_extracted_normalized" in x_column:
                    ax.set_xticks(np.arange(min(x_data), max(x_data)+0.1, 0.1))
                    ax.xaxis.set_major_formatter(plt.FuncFormatter(lambda x, _: '{:.1f}'.format(x)))
                ax.grid(True)
            elif ax == second_row_ax and second_row_ax_used:
                if "Time" in x_column:
                    ax.set_xticks(range(int(min(x_data)), int(max(x_data))+1))
                elif "Excitation" in x_column and "Excitation_BIC_extracted_normalized" not in x_column:
                    ax.set_xticks(np.arange(min(x_data), max(x_data)+0.01, 0.01))
                    ax.xaxis.set_major_formatter(plt.FuncFormatter(lambda x, _: '{:.2f}'.format(x)))
                elif "Excitation_BIC_extracted_normalized" in x_column:
                    ax.set_xticks(np.arange(min(x_data), max(x_data)+0.1, 0.1))
                    ax.xaxis.set_major_formatter(plt.FuncFormatter(lambda x, _: '{:.1f}'.format(x)))
                ax.grid(True)
        
        plt.show()
        

if __name__ == "__main__":
    root = tk.Tk()
    app = GraphPlotterApp(root)
    root.mainloop()
