import numpy as np

def calculate_estimation_lag(target_angle, observed_angle, time_step):
    N = len(target_angle)
    lags = np.arange(-3 * int(1 / time_step), 3 * int(1 / time_step) + 1)  # ±3 seconds range
    R_theta_theta = np.zeros(len(lags))

    target_angle_centered = target_angle - np.mean(target_angle)
    observed_angle_centered = observed_angle - np.mean(observed_angle)

    R_target_target = np.sum(target_angle_centered ** 2)
    R_observed_observed = np.sum(observed_angle_centered ** 2)

    for i, tau in enumerate(lags):
        if tau >= 0:
            R_theta_theta[i] = np.sum(target_angle_centered[:N-tau] * observed_angle_centered[tau:])
        else:
            R_theta_theta[i] = np.sum(target_angle_centered[-tau:] * observed_angle_centered[:N+tau])

    F_tau = R_theta_theta / np.sqrt(R_target_target * R_observed_observed)
    est_lag_index = np.argmax(F_tau)
    est_lag = lags[est_lag_index] * time_step

    return est_lag
