import numpy as np

def sigmoid_function(excitation, sh_REC, xc, excitation_min, excitation_max):
    excitation_norm = (excitation - excitation_min) / (excitation_max - excitation_min)
    sc = (excitation_max - excitation_min) / 2
    yc = (excitation_max + excitation_min) / 2
    return sc * np.tanh(sh_REC * (excitation_norm - xc)) + yc

# 入力を最小値と最大値の間で正規化する関数
def normalize_input(x, min_val, max_val):
    return (x - min_val) / (max_val - min_val)

# 出力を0から1の範囲に正規化する関数
def normalize_output(y, y_min, y_max):
    return (y - y_min) / (y_max - y_min)

# 最終的にmin_valからmax_valの範囲にスケーリングする関数
def scale_output(y_norm, min_val, max_val):
    return y_norm * (max_val - min_val) + min_val

# 非線形単調増加関数の作成
def logarithmic_function(x, min_val, max_val):
    x_norm = normalize_input(x, min_val, max_val)
    y = np.log1p(x_norm)
    y_norm = normalize_output(y, np.log1p(0), np.log1p(1))
    return scale_output(y_norm, min_val, max_val)

def quadratic_function(x, min_val, max_val):
    x_norm = normalize_input(x, min_val, max_val)
    y = x_norm**2
    y_norm = normalize_output(y, 0**2, 1**2)
    return scale_output(y_norm, min_val, max_val)

def cubic_function(x, min_val, max_val):
    x_norm = normalize_input(x, min_val, max_val)
    y = x_norm**3
    y_norm = normalize_output(y, 0**3, 1**3)
    return scale_output(y_norm, min_val, max_val)

def first_order_delay_system(x, min_val, max_val):
    x_norm = normalize_input(x, min_val, max_val)
    y = 1 - np.exp(-5 * x_norm)
    y_norm = normalize_output(y, 1 - np.exp(-5 * 0), 1 - np.exp(-5 * 1))
    return scale_output(y_norm, min_val, max_val)

def tanh(x, min_val, max_val):
    x_norm = normalize_input(x, min_val, max_val)
    y = np.tanh(3 * x_norm)
    y_norm = normalize_output(y, np.tanh(3 * 0), np.tanh(3 * 1))
    return scale_output(y_norm, min_val, max_val)

def sqrt(x, min_val, max_val):
    x_norm = normalize_input(x, min_val, max_val)
    y = np.sqrt(x_norm)
    y_norm = normalize_output(y, np.sqrt(0), np.sqrt(1))
    return scale_output(y_norm, min_val, max_val)

def sine_squared(x, min_val, max_val):
    x_norm = normalize_input(x, min_val, max_val)
    y = np.sin(np.pi * x_norm / 2)**2
    y_norm = normalize_output(y, np.sin(np.pi * 0 / 2)**2, np.sin(np.pi * 1 / 2)**2)
    return scale_output(y_norm, min_val, max_val)