import numpy as np
import math

def calculate_cost(target_angles, observed_angles):
    error = np.rad2deg(target_angles - observed_angles)
    cost = np.average(np.square(error))
    return cost

def generate_gaussian_white_noise(size, mean, std):
    noise = np.random.normal(loc=mean, scale=std, size=size)
    return noise

def update_parameters(n, parameters, parameters_hat, noises, costs, target_angles, observed_angles):
    a = 0.5/20
    beta = 0.01/500
    sigma_1 = np.sqrt(1)
    sigma_2 = np.sqrt(2)
    sigma_3 = np.sqrt(3)
    theta_n = np.array(parameters[n-1])
    theta_n_hat = np.array(parameters_hat[n-1])
    cost = calculate_cost(target_angles, observed_angles)
    v = np.array([generate_gaussian_white_noise(1, mean=0.0, std=sigma_1)[0],
                  generate_gaussian_white_noise(1, mean=0.0, std=sigma_2)[0],
                  generate_gaussian_white_noise(1, mean=0.0, std=sigma_3)[0]])
    sin_v_n = noises[n-1]
    sin_v_n_plus_1 = np.sin(v)
    sin_v_n_plus_1[2] = 0
    sin_v_n_plus_1[1] = 0
    
    if n == 1:
        theta_n_plus_1_hat = theta_n
        theta_n_plus_1 = theta_n_plus_1_hat + a * sin_v_n_plus_1
        theta_n_plus_1 = np.clip(theta_n_plus_1, np.array(parameters[0]) * 0.3, np.array(parameters[0]) * 1.7)
        f_x = 0
    else:
        J_n = cost
        f_x=0
        theta_n_plus_1_hat = theta_n_hat - beta * (2 / a) * J_n * sin_v_n
        theta_n_plus_1_hat = np.clip(theta_n_plus_1_hat, np.array(parameters[0]) * 0.3, np.array(parameters[0]) * 1.7)
        theta_n_plus_1 = theta_n_plus_1_hat + a * sin_v_n_plus_1
        theta_n_plus_1 = np.clip(theta_n_plus_1, np.array(parameters[0]) * 0.3, np.array(parameters[0]) * 1.7)
    
    return theta_n_plus_1_hat, theta_n_plus_1, sin_v_n_plus_1, cost, f_x