import numpy as np

# 三角形メンバーシップ関数の定義
def triangular_mf(x, a, b, c):
    return np.maximum(0, np.minimum((x - a) / (b - a), (c - x) / (c - b)))

# 台形メンバーシップ関数の定義
def trapezoidal_mf(x, a, b, c, d):
    return np.maximum(0, np.minimum((x - a) / (b - a), np.minimum(1, (d - x) / (d - c))))

# メンバーシップ関数の定義
def define_membership_functions():
    error_mfs = {
        'NL': lambda x: np.maximum(trapezoidal_mf(x, -30, -10, -5, -3), (x <= -10).astype(float)),
        'NM': lambda x: triangular_mf(x, -5, -3, -1.5),
        'NS': lambda x: triangular_mf(x, -3, -1.5, -1),
        'Z': lambda x: triangular_mf(x, -1.5, 0, 1.5),
        'PS': lambda x: triangular_mf(x, 1, 1.5, 3),
        'PM': lambda x: triangular_mf(x, 1.5, 3, 5),
        'PL': lambda x: np.maximum(trapezoidal_mf(x, 3, 5, 10, 30), (x >= 10).astype(float))
    }

    waru = 120000
    nl = -5 / waru
    nm = -3 / waru
    ns = -1.5 / waru
    z = 0 / waru
    ps = 1.5 / waru
    pm = 3 / waru
    pl = 5 / waru

    delta_v_values = {
        'NL': nl,
        'NM': nm,
        'NS': ns,
        'Z': z,
        'PS': ps,
        'PM': pm,
        'PL': pl
    }
    
    return error_mfs, delta_v_values

# ファジィ推論とデフュジフィケーションの実装
def fuzzy_inference(error_value, error_mfs, delta_v_values, rules):
    # ファジィファイ
    fuzzified_values = {key: mf(error_value) for key, mf in error_mfs.items()}

    # ルールベース評価
    rule_results = []
    for antecedent, consequent in rules:
        rule_strength = fuzzified_values[antecedent]
        rule_results.append((rule_strength, delta_v_values[consequent]))

    # デフュジフィケーション（重心法）
    numerator = sum([strength * value for strength, value in rule_results])
    denominator = sum([strength for strength, value in rule_results])

    if denominator == 0:
        return 0  # すべてのメンバーシップ関数の値がゼロの場合を処理

    return numerator / denominator

# ルールの定義
def define_rules():
    rules = [
        ('NL', 'PL'),
        ('NM', 'PM'),
        ('NS', 'PS'),
        ('Z', 'Z'),
        ('PS', 'NS'),
        ('PM', 'NM'),
        ('PL', 'NL')
    ]
    return rules

# 使用例
def create_fuzzy_controller():
    error_mfs, delta_v_values = define_membership_functions()
    rules = define_rules()
    return error_mfs, delta_v_values, rules

def Fuzzy_controller(error_mfs, delta_v_values, rules, target_angle, this_angle):
    error_value = this_angle - target_angle
    output_value = fuzzy_inference(error_value, error_mfs, delta_v_values, rules)
    return output_value